#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""
This script describes a multiperiod class to build an object for the
integrated ultra-supercritical power plant and motlen-salt based
thermal energy storage model.
"""

__author__ = "Naresh Susarla"

import pyomo.environ as pyo
import pandas as pd
from collections import deque
import idaes.logger as idaeslog

# IDAES imports
from idaes.apps.grid_integration.multiperiod.multiperiod import MultiPeriodModel

# DISPATCHES imports
from dispatches.case_studies.fossil_case.ultra_supercritical_plant \
    .storage.multiperiod_integrated_storage_usc import (create_usc_model,
                                                        usc_unfix_dof,
                                                        usc_custom_init,
                                                        get_usc_link_variable_pairs)


def create_multiperiod_usc_model(n_time_points=4, pmin=None, pmax=None):
    """
    Create a multi-period usc_mp cycle object. This object contains a pyomo
    model with a block for each time instance.

    n_time_points: Number of time blocks to create
    """
    multiperiod_usc = MultiPeriodModel(
        n_time_points=n_time_points,
        process_model_func=create_usc_model,
        initialization_func=usc_custom_init,
        unfix_dof_func=usc_unfix_dof,
        linking_variable_func=get_usc_link_variable_pairs,
        use_stochastic_build=False,
        outlvl=idaeslog.INFO,
        )

    flowsheet_options={"pmin": pmin,
                        "pmax": pmax}

    # create the multiperiod object
    multiperiod_usc.build_multi_period_model(
        model_data_kwargs={t: flowsheet_options for t in range(n_time_points)},
        flowsheet_options=flowsheet_options,
    )

    return multiperiod_usc


class MultiPeriodUsc:
    def __init__(
        self, model_data=None
    ):
        """
        Arguments:
            horizon::Int64 - number of time points to use for associated multi-period model

        Returns:
            Float64: Value of power output in last time step
        """
        self.multiperiod_usc = None
        self.result_list = []
        self.result_listimp = []
        self.model_data = model_data

    def populate_model(self, blk, horizon):
        """
        Create a integrated ultra-supercritical power plant and molten salt
        thermal energy storage model using the `MultiPeriod` package.

        Arguments:
            blk: this is an empty block passed in from eithe a bidder or tracker

        Returns:
             None
        """
        tank_min = 76000  # in kg
        tank_max = 6739292  # in kg

        if not blk.is_constructed():
            blk.construct()

        multiperiod_usc = create_multiperiod_usc_model(
            n_time_points=horizon,
            pmin=self.model_data.p_min, pmax=self.model_data.p_max
        )
        blk.usc_mp = multiperiod_usc

        active_blks = multiperiod_usc.get_active_process_blocks()
        active_blks[0].fs.previous_salt_inventory_hot.fix(tank_min)
        active_blks[0].fs.previous_salt_inventory_cold.fix(tank_max-tank_min)
        active_blks[0].fs.previous_power.fix(380)

        # create expression that references underlying power variables
        blk.HOUR = pyo.Set(initialize=range(horizon))
        blk.P_T = pyo.Expression(blk.HOUR)
        blk.tot_cost = pyo.Expression(blk.HOUR)
        blk.hot_level = pyo.Expression(blk.HOUR)
        blk.storage_power = pyo.Expression(blk.HOUR)
        blk.plant_power = pyo.Expression(blk.HOUR)
        blk.plant_duty = pyo.Expression(blk.HOUR)
        blk.hxc_salt = pyo.Expression(blk.HOUR)
        blk.hxc_duty = pyo.Expression(blk.HOUR)
        blk.hxc_salt_Tin = pyo.Expression(blk.HOUR)
        blk.hxc_salt_Tout = pyo.Expression(blk.HOUR)
        blk.hxd_salt = pyo.Expression(blk.HOUR)
        blk.hxd_duty = pyo.Expression(blk.HOUR)
        blk.hxd_salt_Tin = pyo.Expression(blk.HOUR)
        blk.hxd_salt_Tout = pyo.Expression(blk.HOUR)
        blk.hxd_steam_Tout = pyo.Expression(blk.HOUR)
        blk.hxd_steam_vfrac = pyo.Expression(blk.HOUR)
        for (t, b) in enumerate(active_blks):
            blk.P_T[t] = b.fs.net_power
            blk.hot_level[t] = b.fs.salt_inventory_hot
            blk.storage_power[t] = ((-1e-6)
                                 * b.fs.es_turbine.work_mechanical[0])
            blk.plant_duty[t] = b.fs.plant_heat_duty[0]
            blk.tot_cost[t] = (
                     b.fs.operating_cost
                     + (b.fs.plant_fixed_operating_cost
                     + b.fs.plant_variable_operating_cost) / (365 * 24)
                )
            blk.plant_power[t] = b.fs.plant_power_out[0]
            blk.hxc_salt[t] = b.fs.hxc.tube_inlet.flow_mass[0]
            blk.hxc_duty[t] = b.fs.hxc.heat_duty[0]
            blk.hxc_salt_Tin[t] = b.fs.hxc.tube_inlet.temperature[0]
            blk.hxc_salt_Tout[t] = b.fs.hxc.tube_outlet.temperature[0]
            blk.hxd_salt[t] = b.fs.hxd.shell_inlet.flow_mass[0]
            blk.hxd_duty[t] = b.fs.hxd.heat_duty[0]
            blk.hxd_salt_Tin[t] = b.fs.hxd.shell_inlet.temperature[0]
            blk.hxd_salt_Tout[t] = b.fs.hxd.shell_outlet.temperature[0]
            blk.hxd_steam_Tout[t] = b.fs.hxd.cold_side.properties_out[0].temperature
            blk.hxd_steam_vfrac[t] = b.fs.hxd.cold_side.properties_out[0].vapor_frac


        self.multiperiod_usc = multiperiod_usc
        return

    @staticmethod
    def update_model(b, implemented_power_output, realized_soc):

        """
        Update `blk` variables using the actual implemented power output.

        Arguments:
            blk: the block that needs to be updated
            realized soc, i.e. the hot salt storage tank level:
            implemented_power_output:

         Returns:
             None
        """
        multiperiod_usc = b.usc_mp
        active_blks = multiperiod_usc.get_active_process_blocks()

        implemented_power = round(implemented_power_output[-1])
        realized_soc = round(realized_soc[-1])
        print("Implemented Power (MPC)", implemented_power)
        print("Realized SOC (MPC)", realized_soc)

        active_blks[0].fs.previous_power.fix(implemented_power)
        active_blks[0].fs.previous_salt_inventory_hot.fix(realized_soc)

        return

    @staticmethod
    def get_last_delivered_power(b, last_implemented_time_step):

        """
        Returns the last delivered power output.

        Arguments:
            blk: the block
            last_implemented_time_step: time index for the last implemented time
                                        step

        Returns:
            Float64: Value of power output in last time step
        """
        # blk = b
        return pyo.value(b.P_T[last_implemented_time_step])

    @staticmethod
    def get_implemented_profile(b, last_implemented_time_step):

        """
        This method gets the implemented variable profiles in the last optimization solve.

        Arguments:
            blk: a Pyomo block
            last_implemented_time_step: time index for the last implemented time step

         Returns:
             profile: the intended profile, {unit: [...]}
        """
        # blk = b
        multiperiod_usc = b.usc_mp
        active_blks = multiperiod_usc.get_active_process_blocks()
        implemented_power_output = deque(
            [
                pyo.value(active_blks[t].fs.net_power)
                for t in range(last_implemented_time_step + 1)
            ]
        )
        realized_soc = deque(
            [
                pyo.value(active_blks[t].fs.salt_inventory_hot)
                for t in range(last_implemented_time_step + 1)
            ]
        )

        return {
            "implemented_power_output": implemented_power_output,
            "realized_soc": realized_soc,
        }

    def record_results(self, blk, date=None, hour=None, **kwargs):

        """
        Record the operations stats for the model.

        Arguments:
            blk:  pyomo block
            date: current simulation date
            hour: current simulation hour

        Returns:
            None

        """
        # blk = b
        df_list = []
        df_listimp = []
        for t in blk.HOUR:
            result_dict = {}
            result_implemented = {}

            result_dict["Date"] = date
            result_dict["Hour"] = hour


            # simulation inputs
            result_dict["Horizon [hr]"] = int(t)

            # model vars
            result_dict["Thermal Power Generated [MW]"] = float(
                round(pyo.value(blk.P_T[t]), 2)
            )
            result_dict["Total Cost [$]"] = float(
                round(pyo.value(blk.tot_cost[t]), 2)
                )
            result_dict["Hot Tank Level [MT]"] = float(
                round(pyo.value(blk.hot_level[t]), 2)
                )
            result_dict["Plant Heat Duty [MWth]"] = float(
                round(pyo.value(blk.plant_duty[t]), 2)
                )
            result_dict["Storage Power [MW]"] = float(
                round(pyo.value(blk.storage_power[t]), 2)
                )
            result_dict["Plant Power [MW]"] = float(
                round(pyo.value(blk.plant_power[t]), 2)
                )
            result_dict["HXC Duty"] = float(
                round(pyo.value(blk.hxc_duty[t] * 1e-6), 2)
                )
            result_dict["HXD Duty"] = float(
                round(pyo.value(blk.hxd_duty[t] * 1e-6), 2)
                )
            result_dict["HXC Salt Flow"] = float(
                round(pyo.value(blk.hxc_salt[t]), 2)
                )
            result_dict["HXD Salt Flow"] = float(
                round(pyo.value(blk.hxd_salt[t]), 2)
                )
            result_dict["HXC Salt T in"] = float(
                round(pyo.value(blk.hxc_salt_Tin[t]), 2)
                )
            result_dict["HXC Salt T out"] = float(
                round(pyo.value(blk.hxc_salt_Tout[t]), 2)
                )
            result_dict["HXD Salt T in"] = float(
                round(pyo.value(blk.hxd_salt_Tin[t]), 2)
                )
            result_dict["HXD Salt T out"] = float(
                round(pyo.value(blk.hxd_salt_Tout[t]), 2)
                )
            result_dict["HXD Steam T out"] = float(
                round(pyo.value(blk.hxd_steam_Tout[t]), 2)
                )
            result_dict["HXD Steam out Vfrac"] = float(
                round(pyo.value(blk.hxd_steam_vfrac[t]), 2)
                )

            if t == 0:
                # simulation inputs
                result_implemented["Date"] = date
                result_implemented["Hour"] = hour
                result_implemented["Horizon [hr]"] = int(t)

                # model vars
                result_implemented["Thermal Power Generated [MW]"] = float(
                    round(pyo.value(blk.P_T[t]), 2)
                )
                result_implemented["Total Cost [$]"] = float(
                    round(pyo.value(blk.tot_cost[t]), 2)
                    )
                result_implemented["Hot Tank Level [MT]"] = float(
                    round(pyo.value(blk.hot_level[t]), 2)
                    )
                result_implemented["Plant Heat Duty [MWth]"] = float(
                    round(pyo.value(blk.plant_duty[t]), 2)
                    )
                result_implemented["Storage Power [MW]"] = float(
                    round(pyo.value(blk.storage_power[t]), 2)
                    )
                result_implemented["Plant Power [MW]"] = float(
                    round(pyo.value(blk.plant_power[t]), 2)
                    )
                result_implemented["HXC Duty"] = float(
                    round(pyo.value(blk.hxc_duty[t] * 1e-6), 2)
                    )
                result_implemented["HXD Duty"] = float(
                    round(pyo.value(blk.hxd_duty[t] * 1e-6), 2)
                    )
                result_implemented["HXC Salt Flow"] = float(
                    round(pyo.value(blk.hxc_salt[t]), 2)
                    )
                result_implemented["HXD Salt Flow"] = float(
                    round(pyo.value(blk.hxd_salt[t]), 2)
                    )
                result_implemented["HXC Salt T in"] = float(
                    round(pyo.value(blk.hxc_salt_Tin[t]), 2)
                    )
                result_implemented["HXC Salt T out"] = float(
                    round(pyo.value(blk.hxc_salt_Tout[t]), 2)
                    )
                result_implemented["HXD Salt T in"] = float(
                    round(pyo.value(blk.hxd_salt_Tin[t]), 2)
                    )
                result_implemented["HXD Salt T out"] = float(
                    round(pyo.value(blk.hxd_salt_Tout[t]), 2)
                    )
                result_implemented["HXD Steam T out"] = float(
                    round(pyo.value(blk.hxd_steam_Tout[t]), 2)
                    )
                result_implemented["HXD Steam out Vfrac"] = float(
                    round(pyo.value(blk.hxd_steam_vfrac[t]), 2)
                    )
            for key in kwargs:
                result_dict[key] = kwargs[key]
                result_implemented[key] = kwargs[key]

            result_df = pd.DataFrame.from_dict(result_dict, orient="index")
            df_list.append(result_df.T)
            result_df2 = pd.DataFrame.from_dict(result_implemented, orient="index")
            df_listimp.append(result_df2.T)

        # append to result list
        self.result_list.append(pd.concat(df_list))
        self.result_listimp.append(pd.concat(df_listimp))

        return

    def write_results(self, path):
        """
        Write the saved results to a csv file.

        Arguments:
            path: the path to write the results.

        Return:
            None
        """

        pd.concat(self.result_list).to_csv(path, index=False)
        pd.concat(self.result_listimp).to_csv("tracking_results.csv")

    @property
    def power_output(self):
        return "P_T"

    @property
    def total_cost(self):
        return ("tot_cost", 1)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

__author__ = "Radhakrishna Tumbalam Gooty"

# General python imports
import numpy as np
import pandas as pd
import logging
from collections import deque

# Pyomo imports
from pyomo.environ import Set, Expression, value

# IDAES imports
from idaes.apps.grid_integration.multiperiod.multiperiod import MultiPeriodModel

# Flowsheet function imports
from dispatches.case_studies.nuclear_case.nuclear_flowsheet import (
    build_ne_flowsheet,
    fix_dof_and_initialize,
)


def get_nuclear_link_variable_pairs(t1, t2):
    """
    This function returns paris of variables that need to be connected across two time periods

    Args:
        t1: current time block
        t2: next time block

    Returns:
        None
    """
    return [
        (t1.fs.h2_tank.tank_holdup[0], t2.fs.h2_tank.tank_holdup_previous[0])
    ]


def unfix_dof(m):
    """
    This function unfixes a few degrees of freedom for optimization

    Args:
        m: object containing the integrated nuclear plant flowsheet

    Returns:
        None
    """
    # Unfix split fractions in the power splitter
    m.fs.np_power_split.split_fraction.unfix()

    # Unfix the holdup_previous and outflow variables
    m.fs.h2_tank.tank_holdup_previous.unfix()
    m.fs.h2_tank.outlet_to_pipeline.flow_mol.unfix()

    return


def create_multiperiod_nuclear_model(
        n_time_points=4,
        h2_demand=0.35,
        demand_type="variable",
        h2_price=4,
    ):
    """
    This function creates a multi-period integrated nuclear flowsheet object. This object contains 
    a pyomo model with a block for each time instance.

    Args:
        n_time_points: Number of time blocks to create

    Returns:
        Object containing multi-period integrated nuclear flowsheet model
    """
    mp_nuclear = MultiPeriodModel(
        n_time_points=n_time_points,
        process_model_func=build_ne_flowsheet,
        linking_variable_func=get_nuclear_link_variable_pairs,
        initialization_func=fix_dof_and_initialize,
        unfix_dof_func=unfix_dof,
        outlvl=logging.WARNING,
    )

    flowsheet_options={
        "np_capacity": 500, 
        "include_turbine": False,
        "pem_capacity": 100,
        "tank_capacity": 5000,
    }

    # create the multiperiod object
    mp_nuclear.build_multi_period_model(
        model_data_kwargs={t: flowsheet_options for t in range(n_time_points)},
        flowsheet_options=flowsheet_options,
        initialization_options={
            "split_frac_grid": 0.95,
            "tank_holdup_previous": 0,
            "flow_mol_to_pipeline": 1,
            "flow_mol_to_turbine": 0,
        },
    )

    mw_h2 = 2.016 * 1e-3  # Molecular mass of hydrogen in kg/mol

    # Define conversion factors
    # TODO: Use pyunits convert instead of defining conversion factors
    MW_to_kW = 1000
    kW_to_MW = 1e-3
    hours_to_s = 3600

    # TODO: Need to find a better way to import costing data.
    # Using a VOM cost of $2.3 per MWh for the nuclear power plant
    # Using a VOM cost of $1.3 per MWh for the PEM electrolyzer
    # Fixed O&M cost for the nuclear power plant is $120,000 per MW-year
    # Fixed O&M cost for the PEM electrolyzer is $47,900 per MW-year
    # Normalized FOM = (120,000 / 8760) = $13.7 per MWh
    # Normalized FOM = (47,900 / 8760) = $5.47 per MWh
    npp_fom = 13.7
    npp_vom = 2.3
    pem_fom = 5.47
    pem_vom = 1.3
    tank_vom = 0.01

    active_process_blks = mp_nuclear.get_active_process_blocks()
    for blk in active_process_blks:
        # Hydrogen demand constraint
        if demand_type == "variable":
            blk.fs.h2_tank.outlet_to_pipeline.flow_mol[0].setub(h2_demand / mw_h2)

        elif demand_type == "fixed":
            blk.fs.h2_tank.outlet_to_pipeline.flow_mol[0].fix(h2_demand / mw_h2)

        # Treating the revenue generated from hydrogen as negative cost
        # To avoid degeneracy, we are adding an operating cost for hydrogen storage
        # FIXME: How do we include the FOM contribution?
        blk.fs.operating_cost = Expression(
            expr=blk.fs.np_power_split.electricity[0] * kW_to_MW * npp_vom +
                blk.fs.pem.electricity[0] * kW_to_MW * pem_vom +
                blk.fs.h2_tank.tank_holdup[0] * mw_h2 * tank_vom -
                blk.fs.h2_tank.outlet_to_pipeline.flow_mol[0] * mw_h2 * hours_to_s * h2_price)

    return mp_nuclear


class MultiPeriodNuclear:
    """
    This class builds an object containing multi-period integrated nuclear flowsheet model
    for the double loop workflow. 

    Args:
        horizon: Length of the time horizon
        pmin: Difference between power output from the nuclear power plant and the PEM capcity
        pmax: Power output from the nuclear power plant
        default_bid_curve: Default bid curve that needs to be used in the bidding process
        generator_name: Name of the generator in RTS - GMLC
    """
    def __init__(self, 
                 model_data):

        # If the default bid curve is not provided use the one below         
        # if default_bid_curve is None:
        #     self.default_bid_curve = {p: 15 for p in np.linspace(pmin, pmax, 5)}
        # else:
        #     self.default_bid_curve = default_bid_curve
        
        self.mp_nuclear = None
        self.result_list = []
        self.model_data = model_data
        self.p_lower = model_data.p_min
        self.p_upper = model_data.p_max
        self.generator = model_data.gen_name

    def populate_model(self, blk, horizon):
        """
        This function creates the nuclear flowsheet model using the `MultiPeriod` class

        Args:
            blk: this is an empty block passed in from either a `Bidder` or `Tracker` object
        
        Returns:
             None
        """
        if not blk.is_constructed():
            blk.construct()

        mp_nuclear = create_multiperiod_nuclear_model(n_time_points=horizon)
        blk.nuclear = mp_nuclear

        active_blks = mp_nuclear.get_active_process_blocks()
        active_blks[0].fs.h2_tank.tank_holdup_previous.fix(0)

        # create expression that references underlying power variables in multi-period rankine
        blk.HOUR = Set(initialize=range(horizon))
        blk.P_T = Expression(blk.HOUR)
        blk.tot_cost = Expression(blk.HOUR)
        
        for (t, b) in enumerate(active_blks):
            blk.P_T[t] = b.fs.np_power_split.np_to_grid_port.electricity[0] * 1e-3
            blk.tot_cost[t] = b.fs.operating_cost

        self.mp_nuclear = mp_nuclear
        return

    @staticmethod
    def update_model(b, implemented_tank_holdup):

        """
        Update `b` variables using the actual implemented power output.

        Args:
            b: the block that needs to be updated
            implemented_tank_holdup: Updated the initial tank holdup

        Returns:
            None
        """
        mp_nuclear = b.nuclear
        active_blks = mp_nuclear.get_active_process_blocks()

        new_init_holdup = round(implemented_tank_holdup[-1])

        # update battery and power output based on implemented values
        active_blks[0].fs.h2_tank.tank_holdup_previous.fix(new_init_holdup)

        return

    @staticmethod
    def get_last_delivered_power(b, last_implemented_time_step):

        """
        Returns the last delivered power output.
        Args:
            b: the block
            last_implemented_time_step: time index for the last implemented time
                                        step
        Returns:
            Float64: Value of power output in last time step
        """
        return value(b.P_T[last_implemented_time_step])

    @staticmethod
    def get_implemented_profile(b, last_implemented_time_step):

        """
        This method gets the implemented variable profiles in the last optimization solve.

        Args:
            b: a Pyomo block
            last_implemented_time_step: time index for the last implemented time step

         Returns:
             profile: the intended profile, {unit: [...]}
        """
        active_blks = b.nuclear.get_active_process_blocks()
        implemented_tank_holdup = deque(
            [
                value(active_blks[t].fs.h2_tank.tank_holdup[0])
                for t in range(last_implemented_time_step + 1)
            ]
        )

        return {
            "implemented_tank_holdup": implemented_tank_holdup
        }

    def record_results(self, blk, date=None, hour=None, **kwargs):

        """
        Record the operations stats for the model.

        Args:
            blk:  pyomo block
            date: current simulation date
            hour: current simulation hour

        Returns:
            None
        """
        active_blks = blk.nuclear.get_active_process_blocks()
        df_list = []
        for t in blk.HOUR:
            result_dict = {
                "Date": date,
                "Hour": hour,
                "Horizon [hr]": int(t),
                "Power to Grid [MW]": float(round(value(blk.P_T[t]), 2)),
                "Power to PEM [MW]": float(round(value(active_blks[t].fs.pem.electricity[0]) * 1e-3, 2)),
                "Initial holdup [kg]": float(round(value(active_blks[t].fs.h2_tank.tank_holdup_previous[0])
                                                   * 2.016 * 1e-3, 2)),
                "Final holdup [kg]": float(round(value(active_blks[t].fs.h2_tank.tank_holdup[0])
                                                 * 2.016 * 1e-3, 2)),
                "Hydrogen Market [kg/hr]": float(round(value(active_blks[t].fs.h2_tank.outlet_to_pipeline.flow_mol[0])
                                                       * 2.016 * 1e-3 * 3600, 2)),
                "Total Cost [$]": float(round(value(blk.tot_cost[t]), 2))}

            for key in kwargs:
                result_dict[key] = kwargs[key]

            result_df = pd.DataFrame.from_dict(result_dict, orient="index")
            df_list.append(result_df.T)

        # append to result list
        self.result_list.append(pd.concat(df_list))

        return

    def write_results(self, path):

        """
        Write the saved results to a csv file.
        
        Args:
            path: the path to write the results.
        Return:
            None
        """

        pd.concat(self.result_list).to_csv(path, index=False)

    @property
    def power_output(self):
        return "P_T"

    @property
    def total_cost(self):
        return ("tot_cost", 1)

    @property
    def pmin(self):
        # return self.generator_data['PMin MW']
        return self.p_lower



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import pytest
# Import objects from pyomo package
from pyomo.environ import ConcreteModel, SolverFactory, Var, Objective
from pyomo.util.check_units import assert_units_consistent

# Import the main FlowsheetBlock from IDAES. The flowsheet block will contain the unit model
from idaes.core import FlowsheetBlock

from dispatches.unit_models import Wind_Power


def test_windpower():
    # Create the ConcreteModel and the FlowsheetBlock, and attach the flowsheet block to it.
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)   # dynamic or ss flowsheet needs to be specified here

    # ((wind m/s, wind degrees from north clockwise, probability), )
    resource_timeseries = dict()
    for time in list(m.fs.config.time.data()):
        resource_timeseries[time] = ((10, 180, 1),)

    wind_config = {'resource_probability_density': resource_timeseries}

    m.fs.unit = Wind_Power(**wind_config)
    assert hasattr(m.fs.unit, "capacity_factor")
    assert hasattr(m.fs.unit, "electricity_out")
    assert isinstance(m.fs.unit.system_capacity, Var)
    assert isinstance(m.fs.unit.electricity, Var)

    m.fs.unit.system_capacity.fix(50000) # kW

    assert_units_consistent(m)

    m.fs.unit.initialize()

    assert m.fs.unit.capacity_factor[0].value == pytest.approx(0.5755, rel=1e-2)
    assert m.fs.unit.electricity_out.electricity[0].value == pytest.approx(28775.06, rel=1e-2)

    solver = SolverFactory('ipopt')
    solver.solve(m.fs)

    assert m.fs.unit.capacity_factor[0].value == pytest.approx(0.5755, rel=1e-2)
    assert m.fs.unit.electricity_out.electricity[0].value <= 28775.06


def test_windpower2():
    # Create the ConcreteModel and the FlowsheetBlock, and attach the flowsheet block to it.
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)   # dynamic or ss flowsheet needs to be specified here

    # ((wind m/s, wind degrees from north clockwise, probability), )
    resource_speed = [10 for _ in list(m.fs.config.time.data())]

    wind_config = {'resource_speed': resource_speed}

    m.fs.unit = Wind_Power(**wind_config)

    m.fs.unit.system_capacity.fix(50000) # kW
    m.fs.unit.initialize()
    m.fs.obj = Objective(expr=-m.fs.unit.electricity_out.electricity[0])

    solver = SolverFactory('ipopt')
    solver.solve(m.fs)

    assert m.fs.unit.electricity_out.electricity[0].value == pytest.approx(30083.39, rel=1e-2)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""
This script uses the IDAES multiperiod class to create a steady state
multiperiod model for the integrated ultra-supercritical power plant and
energy storage system. The purpose of this script is to create a multiperiod
model that can be use for market analysis either using a pricetaker assumption
or in a real-time scenario using the double loop framework. The integrated
storage with ultra-supercritical power plant model is used a steady state model
for creating the multiperiod model.
"""

__author__ = "Naresh Susarla and Soraya Rawlings"

try:
    from importlib import resources  # Python 3.8+
except ImportError:
    import importlib_resources as resources  # Python 3.7
from idaes.core.util import from_json, to_json
import idaes.logger as idaeslog

from pyomo.environ import (NonNegativeReals, Constraint, Var, ConcreteModel)
from idaes.apps.grid_integration.multiperiod.multiperiod import (
    MultiPeriodModel)
from dispatches.case_studies.fossil_case.ultra_supercritical_plant import (
    ultra_supercritical_powerplant as usc)
from dispatches.case_studies.fossil_case.ultra_supercritical_plant.storage import (
    integrated_storage_with_ultrasupercritical_power_plant as usc_w_tes)


def create_usc_model(m=None, pmin=None, pmax=None):

    if m is None:
        m = ConcreteModel(name="Integrated Model")

    # Set bounds for plant power
    min_storage_heat_duty = 10  # in MW
    max_storage_heat_duty = 200  # in MW

    max_power = 436  # in MW
    min_power = int(0.65 * max_power)  # 283 in MW
    if pmin is None:
        pmin = int(0.65 * 436) + 1
    if pmax is None:
        pmax = 436 + 30

    m = usc.build_plant_model(m)

    # Create a flowsheet, add properties, unit models, and arcs
    m = usc_w_tes.create_integrated_model(m, max_power=max_power)

    # Give all the required inputs to the model
    usc_w_tes.set_model_input(m)

    # Add scaling factor
    usc_w_tes.set_scaling_factors(m)

    # Initialize the model with a sequential initialization and custom
    # Add cost correlations
    m = usc_w_tes.build_costing(m)

    # Add bounds
    usc_w_tes.add_bounds(m)
    

    m.fs.plant_min_power_eq = Constraint(
        expr=m.fs.plant_power_out[0] >= min_power
    )
    m.fs.plant_max_power_eq = Constraint(
        expr=m.fs.plant_power_out[0] <= max_power
    )

    m.fs.hxc.heat_duty.setlb(min_storage_heat_duty * 1e6)
    m.fs.hxd.heat_duty.setlb(min_storage_heat_duty * 1e6)

    m.fs.hxc.heat_duty.setub(max_storage_heat_duty * 1e6)
    m.fs.hxd.heat_duty.setub(max_storage_heat_duty * 1e6)

    # Add coupling variables
    m.fs.previous_power = Var(
        domain=NonNegativeReals,
        initialize=300,
        bounds=(pmin, pmax),
        doc="Previous period power (MW)"
        )

    inventory_max = 1e7
    inventory_min = 75000
    tank_max = 6739292           # Units in kg

    m.fs.previous_salt_inventory_hot = Var(
        domain=NonNegativeReals,
        initialize=inventory_min,
        bounds=(0, inventory_max),
        doc="Hot salt at the beginning of the hour (or time period), kg"
        )
    m.fs.salt_inventory_hot = Var(
        domain=NonNegativeReals,
        initialize=inventory_min,
        bounds=(0, inventory_max),
        doc="Hot salt inventory at the end of the hour (or time period), kg"
        )
    m.fs.previous_salt_inventory_cold = Var(
        domain=NonNegativeReals,
        initialize=tank_max-inventory_min,
        bounds=(0, inventory_max),
        doc="Cold salt at the beginning of the hour (or time period), kg"
        )
    m.fs.salt_inventory_cold = Var(
        domain=NonNegativeReals,
        initialize=tank_max-inventory_min,
        bounds=(0, inventory_max),
        doc="Cold salt inventory at the end of the hour (or time period), kg"
        )

    @m.fs.Constraint(doc="Plant ramping down constraint")
    def constraint_ramp_down(b):
        return (
            m.fs.previous_power - 60 <=
            m.fs.plant_power_out[0])

    @m.fs.Constraint(doc="Plant ramping up constraint")
    def constraint_ramp_up(b):
        return (
            m.fs.previous_power + 60 >=
            m.fs.plant_power_out[0])

    @m.fs.Constraint(doc="Inventory balance at the end of the time period")
    def constraint_salt_inventory_hot(b):
        return (
            m.fs.salt_inventory_hot ==
            m.fs.previous_salt_inventory_hot
            + (3600*m.fs.hxc.tube_inlet.flow_mass[0]
               - 3600*m.fs.hxd.shell_inlet.flow_mass[0])
        )

    @m.fs.Constraint(doc="Max salt flow to hxd based on available hot salt")
    def constraint_salt_maxflow_hot(b):
        return (
            3600*m.fs.hxd.shell_inlet.flow_mass[0] <=
            m.fs.previous_salt_inventory_hot
        )

    @m.fs.Constraint(doc="Max salt flow to hxc based on available cold salt")
    def constraint_salt_maxflow_cold(b):
        return (
            3600*m.fs.hxc.tube_inlet.flow_mass[0] <=
            m.fs.previous_salt_inventory_cold
        )

    @m.fs.Constraint(doc="Maximum salt inventory at any time")
    def constraint_salt_inventory(b):
        return (
            m.fs.salt_inventory_hot +
            m.fs.salt_inventory_cold == m.fs.salt_amount)

    return m


def usc_unfix_dof(m):
    # Unfix data
    m.fs.boiler.inlet.flow_mol[0].unfix()

    # Unfix storage system data
    m.fs.ess_hp_split.split_fraction[0, "to_hxc"].unfix()
    m.fs.ess_bfp_split.split_fraction[0, "to_hxd"].unfix()
    for salt_hxc in [m.fs.hxc]:
        salt_hxc.shell_inlet.unfix()
        salt_hxc.tube_inlet.flow_mass.unfix()  # kg/s, 1 DOF
        salt_hxc.area.unfix()  # 1 DOF

    for salt_hxd in [m.fs.hxd]:
        salt_hxd.tube_inlet.unfix()
        salt_hxd.shell_inlet.flow_mass.unfix()  # kg/s, 1 DOF
        salt_hxd.area.unfix()  # 1 DOF

    for unit in [m.fs.cooler]:
        unit.inlet.unfix()
    m.fs.cooler.outlet.enth_mol[0].unfix()  # 1 DOF

    # Fix storage heat exchangers area and salt temperatures
    m.fs.hxc.area.fix(1904)
    m.fs.hxd.area.fix(2830)
    m.fs.hxc.tube_outlet.temperature[0].fix(831)
    m.fs.hxd.shell_inlet.temperature[0].fix(831)
    m.fs.hxd.shell_outlet.temperature[0].fix(513.15)
    


def usc_custom_init(m):

    blk = usc.build_plant_model()
    usc.initialize(blk)

    # Create a flowsheet, add properties, unit models, and arcs
    max_power = 436  # in MW
    min_power = int(0.65 * 436)
    pmin = int(0.65 * 436) + 1
    pmax = 436 + 30
    min_storage_heat_duty = 10  # in MW
    max_storage_heat_duty = 200  # in MW


    blk = usc_w_tes.create_integrated_model(blk, max_power=max_power)

    # Give all the required inputs to the model
    usc_w_tes.set_model_input(blk)

    # Add scaling factor
    usc_w_tes.set_scaling_factors(blk)

    # Initialize the model with a sequential initialization and custom
    # routines
    usc_w_tes.initialize(blk)

    # Add cost correlations
    blk = usc_w_tes.build_costing(blk)

    # Initialize with bounds
    usc_w_tes.initialize_with_costing(blk)

    blk.fs.plant_min_power_eq = Constraint(
        expr=blk.fs.plant_power_out[0] >= min_power
    )
    blk.fs.plant_max_power_eq = Constraint(
        expr=blk.fs.plant_power_out[0] <= max_power
    )

    blk.fs.hxc.heat_duty.setlb(min_storage_heat_duty * 1e6)
    blk.fs.hxd.heat_duty.setlb(min_storage_heat_duty * 1e6)

    blk.fs.hxc.heat_duty.setub(max_storage_heat_duty * 1e6)
    blk.fs.hxd.heat_duty.setub(max_storage_heat_duty * 1e6)

    if pmin is None:
        pmin = int(0.65 * 436) + 1
    if pmax is None:
        pmax = 436 + 30

    # Add coupling variables
    blk.fs.previous_power = Var(
        domain=NonNegativeReals,
        initialize=300,
        bounds=(pmin, pmax),
        doc="Previous period power (MW)"
        )

    inventory_max = 1e7
    inventory_min = 75000
    tank_max = 6739292           # Units in kg

    blk.fs.previous_salt_inventory_hot = Var(
        domain=NonNegativeReals,
        initialize=inventory_min,
        bounds=(0, inventory_max),
        doc="Hot salt at the beginning of the hour (or time period), kg"
        )
    blk.fs.salt_inventory_hot = Var(
        domain=NonNegativeReals,
        initialize=inventory_min,
        bounds=(0, inventory_max),
        doc="Hot salt inventory at the end of the hour (or time period), kg"
        )
    blk.fs.previous_salt_inventory_cold = Var(
        domain=NonNegativeReals,
        initialize=tank_max-inventory_min,
        bounds=(0, inventory_max),
        doc="Cold salt at the beginning of the hour (or time period), kg"
        )
    blk.fs.salt_inventory_cold = Var(
        domain=NonNegativeReals,
        initialize=tank_max-inventory_min,
        bounds=(0, inventory_max),
        doc="Cold salt inventory at the end of the hour (or time period), kg"
        )

    @blk.fs.Constraint(doc="Plant ramping down constraint")
    def constraint_ramp_down(b):
        return (
            blk.fs.previous_power - 60 <=
            blk.fs.plant_power_out[0])

    @blk.fs.Constraint(doc="Plant ramping up constraint")
    def constraint_ramp_up(b):
        return (
            blk.fs.previous_power + 60 >=
            blk.fs.plant_power_out[0])

    @blk.fs.Constraint(doc="Inventory balance at the end of the time period")
    def constraint_salt_inventory_hot(b):
        return (
            blk.fs.salt_inventory_hot ==
            blk.fs.previous_salt_inventory_hot
            + (3600*blk.fs.hxc.tube_inlet.flow_mass[0]
               - 3600*blk.fs.hxd.shell_inlet.flow_mass[0])
        )

    @blk.fs.Constraint(doc="Max salt flow to hxd based on available hot salt")
    def constraint_salt_maxflow_hot(b):
        return (
            3600*blk.fs.hxd.shell_inlet.flow_mass[0] <=
            blk.fs.previous_salt_inventory_hot
        )

    @blk.fs.Constraint(doc="Max salt flow to hxc based on available cold salt")
    def constraint_salt_maxflow_cold(b):
        return (
            3600*blk.fs.hxc.tube_inlet.flow_mass[0] <=
            blk.fs.previous_salt_inventory_cold
        )

    @blk.fs.Constraint(doc="Maximum salt inventory at any time")
    def constraint_salt_inventory(b):
        return (
            blk.fs.salt_inventory_hot +
            blk.fs.salt_inventory_cold == blk.fs.salt_amount)

    init_model = to_json(blk, return_dict=True)
    from_json(m, sd=init_model)

    return


# The tank level and power output are linked between the contiguous time periods
def get_usc_link_variable_pairs(b1, b2):
    """
        b1: current time block
        b2: next time block
    """
    return [(b1.fs.salt_inventory_hot,
             b2.fs.previous_salt_inventory_hot),
            (b1.fs.plant_power_out[0],
             b2.fs.previous_power)]


# The tank level at the end of the last time period must be the same as at the
# beginning of the first time period
def get_usc_periodic_variable_pairs(b1, b2):
    """
        b1: final time block
        b2: first time block
    """
    # return
    return [(b1.fs.salt_inventory_hot,
             b2.fs.previous_salt_inventory_hot)]

# Create the multiperiod model object. You can pass arguments to your
# "process_model_func" for each time period using a dict of dicts as
# shown here.  In this case, it is setting up empty dictionaries for
# each time period.


def create_multiperiod_usc_model(n_time_points=4, pmin=None, pmax=None):
    """
    Create a multi-period usc_mp cycle object. This object contains a pyomo
    model with a block for each time instance.

    n_time_points: Number of time blocks to create
    """
    multiperiod_usc = MultiPeriodModel(
        n_time_points=n_time_points,
        process_model_func=create_usc_model,
        initialization_func=usc_custom_init,
        unfix_dof_func=usc_unfix_dof,
        linking_variable_func=get_usc_link_variable_pairs,
        flowsheet_options={"pmin": pmin,
                           "pmax": pmax},
        use_stochastic_build=True,
        outlvl=idaeslog.INFO,
        )

    return multiperiod_usc



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
import pandas as pd
import pyomo.environ as pyo
from idaes.apps.grid_integration.multiperiod.multiperiod import MultiPeriodModel
from dispatches.case_studies.renewables_case.RE_flowsheet import *
from dispatches.case_studies.renewables_case.load_parameters import *
from dispatches.case_studies.renewables_case.load_parameters import df


def wind_battery_variable_pairs(m1, m2):
    """
    This function links together unit model state variables from one timestep to the next.

    The battery's state of charge and energy throughput values need to be consistent across time blocks.

    Args:
        m1: current time block model
        m2: next time block model
    """
    pairs = [
        (m1.fs.battery.state_of_charge[0], m2.fs.battery.initial_state_of_charge),
        (m1.fs.battery.energy_throughput[0], m2.fs.battery.initial_energy_throughput),
        (m1.fs.battery.nameplate_power, m2.fs.battery.nameplate_power)
    ]
    return pairs


def wind_battery_periodic_variable_pairs(m1, m2):
    """
    The final battery storage of charge must be the same as in the intial timestep. 

    Args:
        m1: final time block model
        m2: first time block model
    """
    pairs = [(m1.fs.battery.state_of_charge[0], m2.fs.battery.initial_state_of_charge),
             (m1.fs.battery.nameplate_power, m2.fs.battery.nameplate_power)]
    return pairs


def wind_battery_om_costs(m):
    """
    Add unit fixed and variable operating costs as parameters for the unit model m
    """
    m.fs.windpower.op_cost = pyo.Param(
        initialize=wind_op_cost, doc="fixed cost of operating wind plant $10/kW-yr"
    )
    m.fs.windpower.op_total_cost = Expression(
        expr=m.fs.windpower.system_capacity * m.fs.windpower.op_cost / 8760,
        doc="total fixed cost of wind in $/hr",
    )
    m.fs.battery.var_cost = pyo.Expression(
        expr=m.fs.battery.degradation_rate * (m.fs.battery.energy_throughput[0] - m.fs.battery.initial_energy_throughput) * batt_rep_cost_kwh,
        doc="variable operating of the battery $/kWh"
    )


def initialize_mp(m, verbose=False):
    """
    Initializing the flowsheet is done starting with the wind model and propagating the solved initial state to the splitter and battery.

    The splitter is initialized with no flow to the battery so all electricity flows to the grid, which makes the initialization of all
    unit models downstream of the wind plant independent of its time-varying electricity production. This initialzation function can
    then be repeated for all timesteps within a dynamic analysis.

    Args:
        m: model
        verbose:
    """
    outlvl = idaeslog.INFO if verbose else idaeslog.WARNING

    m.fs.windpower.initialize(outlvl=outlvl)

    propagate_state(m.fs.wind_to_splitter)
    m.fs.splitter.battery_elec[0].fix(1)
    m.fs.splitter.initialize()
    m.fs.splitter.battery_elec[0].unfix()
    if verbose:
        m.fs.splitter.report(dof=True)

    propagate_state(m.fs.splitter_to_battery)
    m.fs.battery.elec_in[0].fix()
    m.fs.battery.elec_out[0].fix(value(m.fs.battery.elec_in[0]))
    m.fs.battery.initialize(outlvl=outlvl)
    m.fs.battery.elec_in[0].unfix()
    m.fs.battery.elec_out[0].unfix()
    if verbose:
        m.fs.battery.report(dof=True)


def wind_battery_model(wind_resource_config, input_params, verbose=False):
    """
    Creates an initialized flowsheet model for a single time step with operating, size and cost components
    
    First, the model is created using the input_params and wind_resource_config
    Second, the model is initialized so that it solves and its values are internally consistent
    Third, battery ramp constraints and operating cost components are added

    Args:
        wind_resource_config: wind resource for the time step
        input_params: size and operation parameters. Required keys: `wind_mw` and `batt_mw`
        verbose:
    """
    m = create_model(
        input_params['wind_mw'],
        None,
        input_params['batt_mw'],
        None,
        None,
        None,
        wind_resource_config=wind_resource_config
    )

    m.fs.battery.initial_state_of_charge.fix(0)
    m.fs.battery.initial_energy_throughput.fix(0)
    initialize_mp(m, verbose=verbose)

    wind_battery_om_costs(m)
    m.fs.battery.initial_state_of_charge.unfix()
    m.fs.battery.initial_energy_throughput.unfix()

    batt = m.fs.battery

    batt.energy_down_ramp = pyo.Constraint(
        expr=batt.initial_state_of_charge - batt.state_of_charge[0] <= battery_ramp_rate)
    batt.energy_up_ramp = pyo.Constraint(
        expr=batt.state_of_charge[0] - batt.initial_state_of_charge <= battery_ramp_rate)

    return m


def wind_battery_mp_block(wind_resource_config, input_params, verbose=False):
    """
    Wrapper of `wind_battery_model` for creating the process model per time point for the MultiPeriod model.
    Uses cloning of the Pyomo model in order to reduce runtime. 
    The clone is reinitialized with the `wind_resource_config` for the given time point, which only required modifying
    the windpower and the splitter, as the rest of the units have no flow and therefore is unaffected by wind resource changes.

    Args:
        wind_resource_config: dictionary with `resource_speed` for the time step
        input_params: size and operation parameters. Required keys: `wind_mw`, `batt_mw`
        verbose:
    """
    if 'pyo_model' not in input_params.keys():
        input_params['pyo_model'] = wind_battery_model(wind_resource_config, input_params, verbose=verbose)
    m = input_params['pyo_model'].clone()
    if 'resource_speed' in wind_resource_config.keys():
        m.fs.windpower.config.resource_speed = wind_resource_config['resource_speed']
    elif 'capacity_factor' in wind_resource_config.keys():
        m.fs.windpower.config.capacity_factor = wind_resource_config['capacity_factor']
    else:
        raise ValueError(f"`wind_resource_config` dict must contain either 'resource_speed' or 'capacity_factor' values")
    m.fs.windpower.setup_resource()
    return m


def wind_battery_optimize(n_time_points, input_params, verbose=False):
    """
    The main function for optimizing the flowsheet's design and operating variables for Net Present Value. 

    Creates the MultiPeriodModel and adds the size and operating constraints in addition to the Net Present Value Objective.
    The NPV is a function of the capital costs, the electricity market profit and the capital recovery factor.
    The operating decisions and state evolution of the unit models and the flowsheet as a whole form the constraints of the Linear Program.

    Required input parameters include:
        `wind_mw`: initial guess of the wind size
        `wind_mw_ub`: upper bound of wind size
        `batt_mw`: initial guess of the battery size
        `wind_resource`: dictionary of wind resource configs for each time point
        `DA_LMPs`: LMPs for each time point
        `design_opt`: true to optimize design, else sizes are fixed at initial guess sizes
        `extant_wind`: if true, fix wind size to initial size and do not add wind capital cost to NPV

    Args:
        n_time_points: number of periods in MultiPeriod model
        input_params: 
        verbose: print all logging and outputs from unit models, initialization, solvers, etc
        plot: plot the operating variables time series
    """
    mp_wind_battery = MultiPeriodModel(
        n_time_points=n_time_points,
        process_model_func=partial(wind_battery_mp_block, input_params=input_params, verbose=verbose),
        linking_variable_func=wind_battery_variable_pairs,
        periodic_variable_func=wind_battery_periodic_variable_pairs,
    )

    mp_wind_battery.build_multi_period_model(input_params['wind_resource'])

    m = mp_wind_battery.pyomo_model
    blks = mp_wind_battery.get_active_process_blocks()
    blks[0].fs.battery.initial_state_of_charge.fix(0)
    blks[0].fs.battery.initial_energy_throughput.fix(0)

    m.wind_system_capacity = Var(domain=NonNegativeReals, initialize=input_params['wind_mw'] * 1e3, units=pyunits.kW, bounds=(0, input_params['wind_mw_ub'] * 1e3))
    m.battery_system_capacity = Var(domain=NonNegativeReals, initialize=input_params['batt_mw'] * 1e3, units=pyunits.kW)
    
    if input_params['design_opt']:
        for blk in blks:
            if not input_params['extant_wind']:
                blk.fs.windpower.system_capacity.unfix()
            blk.fs.battery.nameplate_power.unfix()
    
    m.wind_max_p = Constraint(mp_wind_battery.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.windpower.system_capacity <= m.wind_system_capacity)
    m.battery_max_p = Constraint(mp_wind_battery.pyomo_model.TIME, rule=lambda b, t: blks[t].fs.battery.nameplate_power <= m.battery_system_capacity)

    # add market data for each block
    for blk in blks:
        blk_wind = blk.fs.windpower
        blk_battery = blk.fs.battery
        
        # add operating costs
        blk_wind.op_total_cost = Expression(
            expr=blk_wind.system_capacity * blk_wind.op_cost / 8760
        )

        blk.lmp_signal = pyo.Param(default=0, mutable=True)
        blk.revenue = (
            blk.lmp_signal * (blk.fs.splitter.grid_elec[0] + blk_battery.elec_out[0])
        )
        blk.profit = pyo.Expression(expr=blk.revenue 
                                         - blk_wind.op_total_cost
                                         - blk_battery.var_cost)

    for (i, blk) in enumerate(blks):
        blk.lmp_signal.set_value(input_params['DA_LMPs'][i] * 1e-3) 
    
    m.wind_cap_cost = pyo.Param(default=wind_cap_cost, mutable=True)
    if input_params['extant_wind']:
        m.wind_cap_cost.set_value(0.0)
    m.batt_cap_cost = pyo.Param(default=batt_cap_cost, mutable=True)

    n_weeks = n_time_points / (7 * 24)
    m.annual_revenue = Expression(expr=sum([blk.profit for blk in blks]) * 52 / n_weeks)
    m.NPV = Expression(
        expr=-(
            m.wind_cap_cost * m.wind_system_capacity
            + m.batt_cap_cost * m.battery_system_capacity
        )
        + PA * m.annual_revenue
    )
    m.obj = pyo.Objective(expr=-m.NPV * 1e-5)

    opt = pyo.SolverFactory("cbc")
    opt.solve(m, tee=verbose)

    return mp_wind_battery


def record_results(mp_wind_battery):

    m = mp_wind_battery.pyomo_model

    batt_to_grid = []
    wind_to_grid = []
    wind_to_batt = []
    wind_gen = []
    soc = []
    elec_revenue = []

    blks = mp_wind_battery.get_active_process_blocks()
    n_time_points = len(blks)
    soc = [
        pyo.value(blks[i].fs.battery.state_of_charge[0]) for i in range(n_time_points)
    ]
    wind_gen = [
        pyo.value(blks[i].fs.windpower.electricity[0]) * 1e-3 for i in range(n_time_points)
    ]
    batt_to_grid = [
        pyo.value(blks[i].fs.battery.elec_out[0]) * 1e-3 for i in range(n_time_points)
    ]
    wind_to_grid = [pyo.value(blks[i].fs.splitter.grid_elec[0]) * 1e-3 for i in range(n_time_points)]
    wind_to_batt = [
        pyo.value(blks[i].fs.battery.elec_in[0]) * 1e-3 for i in range(n_time_points)
    ]
    elec_revenue = [pyo.value(blks[i].profit) for i in range(n_time_points)]
    lmp = [pyo.value(blks[i].lmp_signal) * 1e3 for i in range(n_time_points)]

    wind_cap = value(m.wind_system_capacity) * 1e-3
    batt_cap = value(m.battery_system_capacity) * 1e-3

    annual_revenue = value(m.annual_revenue)
    npv = value(m.NPV)

    print("wind mw", wind_cap)
    print("batt mw", batt_cap)
    print("elec rev", sum(elec_revenue))
    print("annual rev", annual_revenue)
    print("npv", npv)

    return (
        soc,
        wind_gen,
        batt_to_grid,
        wind_to_grid,
        wind_to_batt,
        elec_revenue,
        lmp,
        wind_cap,
        batt_cap,
        annual_revenue,
        npv,
    )


def plot_results(
    soc,
    wind_gen,
    batt_to_grid,
    wind_to_grid,
    wind_to_batt,
    elec_revenue,
    lmp,
    wind_cap,
    batt_cap,
    annual_revenue,
    npv,
):

    hours = [t for t in range(len(soc))]
    fig, ax1 = plt.subplots(2, 1, figsize=(12, 8))
    plt.suptitle(
        f"Optimal NPV ${round(npv * 1e-6)}mil from {round(batt_cap, 2)} MW Battery"
    )

    # color = 'tab:green'
    ax1[0].set_xlabel("Hour")
    ax1[0].set_ylabel(
        "MW",
    )
    ax1[0].step(hours, wind_gen, label="Wind Generation [MW]")
    ax1[0].step(hours, wind_to_grid, label="Wind to Grid [MW]")
    ax1[0].step(hours, wind_to_batt, label="Wind to Batt [MW]")
    ax1[0].step(hours, batt_to_grid, label="Batt to Grid [MW]")
    ax1[0].tick_params(
        axis="y",
    )
    ax1[0].legend()
    ax1[0].grid(visible=True, which="major", color="k", linestyle="--", alpha=0.2)
    ax1[0].minorticks_on()
    ax1[0].grid(visible=True, which="minor", color="k", linestyle="--", alpha=0.2)

    ax2 = ax1[0].twinx()
    color = "k"
    ax2.set_ylabel("LMP [$/MWh]", color=color)
    ax2.plot(hours, lmp, color=color)
    ax2.tick_params(axis="y", labelcolor=color)

    ax2 = ax1[1].twinx()
    color = "k"
    ax2.set_ylabel("LMP [$/MWh]", color=color)
    ax2.plot(hours, lmp, color=color)
    ax2.tick_params(axis="y", labelcolor=color)

    ax1[1].set_xlabel("Hour")
    ax1[1].step(hours, elec_revenue, label="Elec rev")
    ax1[1].step(hours, np.cumsum(elec_revenue), label="Elec rev cumulative")
    ax1[1].legend()
    ax1[1].grid(visible=True, which="major", color="k", linestyle="--", alpha=0.2)
    ax1[1].minorticks_on()
    ax1[1].grid(visible=True, which="minor", color="k", linestyle="--", alpha=0.2)
    # plt.show()

    return ax1, ax2


if __name__ == "__main__":
    mp_wind_battery = wind_battery_optimize(n_time_points=6 * 24, input_params=default_input_params)
    soc, wind_gen, batt_to_grid, wind_to_grid, wind_to_batt, elec_revenue, lmp, wind_cap, batt_cap, annual_revenue, npv = record_results(mp_wind_battery)
    ax1, ax2 = plot_results(soc, wind_gen, batt_to_grid, wind_to_grid, wind_to_batt, elec_revenue, lmp, wind_cap, batt_cap, annual_revenue, npv)
    # plt.show()



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""This is a Generalized Disjunctive Programming model for the
conceptual design of an ultra supercritical coal-fired power plant
integrated with a discharge storage system

"""

__author__ = "Naresh Susarla and Soraya Rawlings"

import logging

# Import Python libraries
from math import pi
from IPython import embed

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.environ import (Block, Param, Constraint, Objective,
                           TransformationFactory, SolverFactory,
                           Expression, value, log, exp, Var)
from pyomo.environ import units as pyunits
from pyomo.network import Arc
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.gdp import Disjunct, Disjunction
from pyomo.network.plugins import expand_arcs
from pyomo.contrib.fbbt.fbbt import  _prop_bnds_root_to_leaf_map
from pyomo.core.expr.numeric_expr import ExternalFunctionExpression

# Import IDAES libraries
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.initialization import propagate_state
from idaes.core.solvers.get_solver import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core import UnitModelCostingBlock
from idaes.models.unit_models import HeatExchanger
from idaes.models.unit_models.heat_exchanger import delta_temperature_underwood_callback
from idaes.models_extra.power_generation.unit_models.helm import (HelmTurbineStage,
                                                                  HelmSplitter)
from idaes.models.costing.SSLW import (
    SSLWCosting,
    SSLWCostingData
)
from idaes.core.util.exceptions import ConfigurationError

# Import ultra supercritical power plant model
from dispatches.case_studies.fossil_case.ultra_supercritical_plant import (
    ultra_supercritical_powerplant as usc)

# Import properties package for Solar salt
from dispatches.properties import solarsalt_properties


scaling_obj = 1

def create_discharge_model(m, add_efficiency=None, power_max=None):
    """Create flowsheet and add unit models.

    """

    # Create a block to add charge storage model
    m.fs.discharge = Block()

    # Add model data
    _add_data(m)

    # Add Solar salt properties
    m.fs.solar_salt_properties = solarsalt_properties.SolarsaltParameterBlock()

    ###########################################################################
    #  Add unit models
    ###########################################################################

    # Declare splitter to divert condensate to discharge storage heat
    # exchanger
    m.fs.discharge.es_split = HelmSplitter(
        property_package=m.fs.prop_water,
        outlet_list=["to_fwh", "to_hxd"],
    )

    # Declare discharge storage heat exchanger
    m.fs.discharge.hxd = HeatExchanger(
        delta_temperature_callback=delta_temperature_underwood_callback,
        hot_side_name="shell",
        cold_side_name="tube",
        shell={"property_package": m.fs.solar_salt_properties},
        tube={"property_package": m.fs.prop_water},
    )

    # Declare turbine for storage system
    m.fs.discharge.es_turbine = HelmTurbineStage(
        property_package=m.fs.prop_water
    )

    ###########################################################################
    #  Declare disjuncts
    ###########################################################################
    # Disjunction 1 for the sink of discharge HX consists of 2 disjuncts:
    #   1. condpump_source_disjunct ======> condensate from condenser pump
    #   2. fwh4_source_disjunct     ======> condensate from feed water heater 4
    #   3. booster_source_disjunct  ======> condensate from booster pump
    #   4. bfp_source_disjunct      ======> condensate from boiler feed water pump
    #   5. fwh9_source_disjunct     ======> condensate from feed water heater 9

    # Declare disjuncts in disjunction 1
    m.fs.discharge.condpump_source_disjunct = Disjunct(
        rule=condpump_source_disjunct_equations)
    m.fs.discharge.fwh4_source_disjunct = Disjunct(
        rule=fwh4_source_disjunct_equations)
    m.fs.discharge.booster_source_disjunct = Disjunct(
        rule=booster_source_disjunct_equations)
    m.fs.discharge.bfp_source_disjunct = Disjunct(
        rule=bfp_source_disjunct_equations)
    m.fs.discharge.fwh9_source_disjunct = Disjunct(
        rule=fwh9_source_disjunct_equations)

    ###########################################################################
    # Add constraints
    ###########################################################################
    _make_constraints(m, add_efficiency=add_efficiency, power_max=power_max)

    _solar_salt_ohtc_calculation(m)

    ###########################################################################
    #  Create stream arcs
    ###########################################################################

    _create_arcs(m)
    TransformationFactory("network.expand_arcs").apply_to(m.fs.discharge)

    return m


def _add_data(m):
    """Add data to the model
    """

    # Add Chemical Engineering cost index for 2019
    m.CE_index = 607.5

    # Add operating hours
    m.fs.discharge.hours_per_day = pyo.Param(
        initialize=6,
        doc='Number of hours of charging per day'
    )

    # Define number of years over which the costs are annualized
    m.fs.discharge.num_of_years = pyo.Param(
        initialize=30,
        doc='Number of years for cost annualization')

    # Add data to compute overall heat transfer coefficient for the
    # Solar salt storage heat exchanger using the Sieder-Tate
    # correlation. Parameters for tube diameter and thickness assumed
    # from the data in (2017) He et al., Energy Procedia 105, 980-985
    m.fs.discharge.data_hxd = {
        'tube_inner_dia': 0.032,
        'tube_outer_dia': 0.036,
        'k_steel': 21.5,
        'number_tubes': 20,
        'shell_inner_dia': 1
    }
    m.fs.discharge.hxd_tube_inner_dia = pyo.Param(
        initialize=m.fs.discharge.data_hxd['tube_inner_dia'],
        doc='Tube inner diameter in m')
    m.fs.discharge.hxd_tube_outer_dia = pyo.Param(
        initialize=m.fs.discharge.data_hxd['tube_outer_dia'],
        doc='Tube outer diameter in m')
    m.fs.discharge.hxd_k_steel = pyo.Param(
        initialize=m.fs.discharge.data_hxd['k_steel'],
        doc='Thermal conductivity of steel in W/m.K')
    m.fs.discharge.hxd_n_tubes = pyo.Param(
        initialize=m.fs.discharge.data_hxd['number_tubes'],
        doc='Number of tubes')
    m.fs.discharge.hxd_shell_inner_dia = pyo.Param(
        initialize=m.fs.discharge.data_hxd['shell_inner_dia'],
        doc='Shell inner diameter in m')

    # Calculate sectional area of storage heat exchanger
    m.fs.discharge.hxd_tube_cs_area = pyo.Expression(
        expr=(pi / 4) *
        (m.fs.discharge.hxd_tube_inner_dia**2),
        doc="Tube inside cross sectional area in m2")
    m.fs.discharge.hxd_tube_out_area = pyo.Expression(
        expr=(pi / 4) *
        (m.fs.discharge.hxd_tube_outer_dia**2),
        doc="Tube cross sectional area including thickness in m2")
    m.fs.discharge.hxd_shell_eff_area = pyo.Expression(
        expr=(
            (pi / 4) *
            m.fs.discharge.hxd_shell_inner_dia**2 -
            m.fs.discharge.hxd_n_tubes *
            m.fs.discharge.hxd_tube_out_area
        ),
        doc="Effective shell cross sectional area in m2")

    m.fs.discharge.hxd_tube_dia_ratio = (
        m.fs.discharge.hxd_tube_outer_dia / m.fs.discharge.hxd_tube_inner_dia)
    m.fs.discharge.hxd_log_tube_dia_ratio = log(m.fs.discharge.hxd_tube_dia_ratio)

    # Add fuel cost data
    m.data_cost = {
        'coal_price': 2.11e-9,
    }
    m.fs.discharge.coal_price = pyo.Param(
        initialize=m.data_cost['coal_price'],
        doc='Coal price based on HHV in $/J')

    # Add parameters to calculate the Solar salt pump costing. Since
    # the unit is not explicitly modeled, the IDAES cost method is not
    # used for this equipment.  The primary purpose of the salt pump
    # is to move the molten salt without changing the pressure. Thus,
    # the pressure head is computed assuming that the salt is moved on
    # an average of 5m linear distance.
    m.data_salt_pump = {
        'FT': 1.5,
        'FM': 2.0,
        'head': 3.281*5,
        'motor_FT': 1,
        'nm': 1
    }
    m.fs.discharge.spump_FT = pyo.Param(
        initialize=m.data_salt_pump['FT'],
        doc='Pump Type Factor for vertical split case')
    m.fs.discharge.spump_FM = pyo.Param(
        initialize=m.data_salt_pump['FM'],
        doc='Pump Material Factor Stainless Steel')
    m.fs.discharge.spump_head = pyo.Param(
        initialize=m.data_salt_pump['head'],
        doc='Pump Head 5m in ft.')
    m.fs.discharge.spump_motorFT = pyo.Param(
        initialize=m.data_salt_pump['motor_FT'],
        doc='Motor Shaft Type Factor')
    m.fs.discharge.spump_nm = pyo.Param(
        initialize=m.data_salt_pump['nm'],
        doc='Motor Shaft Type Factor')


def _make_constraints(m, add_efficiency=None, power_max=None):
    """Declare constraints for the discharge model

    """

    # Add a constraint to storage turbine to ensure that the outlet
    # temperature is at the saturation temperature
    @m.fs.discharge.es_turbine.Constraint(
        m.fs.time,
        doc="Turbine outlet should be a saturated steam")
    def constraint_esturbine_temperature_out(b, t):
        return (
            b.control_volume.properties_out[t].temperature ==
            b.control_volume.properties_out[t].temperature_sat + 1
        )

    m.fs.net_power = pyo.Expression(
        expr=(m.fs.plant_power_out[0]
              + (m.fs.discharge.es_turbine.control_volume.work[0] * (-1e-6)))
    )

    m.fs.boiler_efficiency = pyo.Var(initialize=0.9,
                                     bounds=(0, 1),
                                     doc="Boiler efficiency")
    m.fs.boiler_efficiency_eq = pyo.Constraint(
        expr=m.fs.boiler_efficiency == (
            0.2143 *
            (m.fs.net_power / power_max) +
            0.7357
        ),
        doc="Boiler efficiency in fraction"
    )
    m.fs.coal_heat_duty = pyo.Var(
        initialize=1000,
        bounds=(0, 1e5),
        doc="Coal heat duty supplied to boiler in MW")

    if add_efficiency:
        m.fs.coal_heat_duty_eq = pyo.Constraint(
            expr=m.fs.coal_heat_duty *
            m.fs.boiler_efficiency ==
            m.fs.plant_heat_duty[0]
        )
    else:
        m.fs.coal_heat_duty_eq = pyo.Constraint(
            expr=m.fs.coal_heat_duty == m.fs.plant_heat_duty[0]
        )

    m.fs.cycle_efficiency = pyo.Var(initialize=0.4,
                                    bounds=(0, 1),
                                    doc="Cycle efficiency")
    m.fs.cycle_efficiency_eq = pyo.Constraint(
        expr=(
            m.fs.cycle_efficiency *
            m.fs.coal_heat_duty
        ) == m.fs.net_power,
        doc="Cycle efficiency"
    )


def _solar_salt_ohtc_calculation(m):
    """Block of equations to compute overall heat transfer coefficient for
    Solar salt heat exchanger

    """

    # Calculate Reynolds number for the salt
    m.fs.discharge.hxd.salt_reynolds_number = pyo.Expression(
        expr=(
            (m.fs.discharge.hxd.shell_inlet.flow_mass[0] *
             m.fs.discharge.hxd_tube_outer_dia) /
            (m.fs.discharge.hxd_shell_eff_area *
             m.fs.discharge.hxd.hot_side.properties_in[0].visc_d_phase["Liq"])
        ),
        doc="Salt Reynolds Number")

    # Calculate Prandtl number for the salt
    m.fs.discharge.hxd.salt_prandtl_number = pyo.Expression(
        expr=(
            m.fs.discharge.hxd.hot_side.properties_in[0].cp_mass["Liq"] *
            m.fs.discharge.hxd.hot_side.properties_in[0].visc_d_phase["Liq"] /
            m.fs.discharge.hxd.hot_side.properties_in[0].therm_cond_phase["Liq"]
        ),
        doc="Salt Prandtl Number")

    # Calculate Prandtl Wall number for the salt
    m.fs.discharge.hxd.salt_prandtl_wall = pyo.Expression(
        expr=(
            m.fs.discharge.hxd.hot_side.properties_out[0].cp_mass["Liq"] *
            m.fs.discharge.hxd.hot_side.properties_out[0].visc_d_phase["Liq"] /
            m.fs.discharge.hxd.hot_side.properties_out[0].therm_cond_phase["Liq"]
        ),
        doc="Salt Prandtl Number at wall")

    # Calculate Nusselt number for the salt
    m.fs.discharge.hxd.salt_nusselt_number = pyo.Expression(
        expr=(
            0.35 *
            (m.fs.discharge.hxd.salt_reynolds_number**0.6) *
            (m.fs.discharge.hxd.salt_prandtl_number**0.4) *
            ((m.fs.discharge.hxd.salt_prandtl_number /
              m.fs.discharge.hxd.salt_prandtl_wall) ** 0.25) *
            (2**0.2)
        ),
        doc="Salt Nusslet Number from 2019, App Ener (233-234), 126")

    # Calculate Reynolds number for the steam
    m.fs.discharge.hxd.steam_reynolds_number = pyo.Expression(
        expr=(
            m.fs.discharge.hxd.tube_inlet.flow_mol[0] *
            m.fs.discharge.hxd.cold_side.properties_in[0].mw *
            m.fs.discharge.hxd_tube_inner_dia /
            (m.fs.discharge.hxd_tube_cs_area *
             m.fs.discharge.hxd_n_tubes *
             m.fs.discharge.hxd.cold_side.properties_in[0].visc_d_phase["Vap"])
        ),
        doc="Steam Reynolds Number")

    # Calculate Reynolds number for the steam
    m.fs.discharge.hxd.steam_prandtl_number = pyo.Expression(
        expr=(
            (m.fs.discharge.hxd.cold_side.properties_in[0].cp_mol /
             m.fs.discharge.hxd.cold_side.properties_in[0].mw) *
            m.fs.discharge.hxd.cold_side.
            properties_in[0].visc_d_phase["Vap"] /
            m.fs.discharge.hxd.cold_side.properties_in[0].therm_cond_phase["Vap"]
        ),
        doc="Steam Prandtl Number")

    # Calculate Reynolds number for the steam
    m.fs.discharge.hxd.steam_nusselt_number = pyo.Expression(
        expr=(
            0.023 *
            (m.fs.discharge.hxd.steam_reynolds_number**0.8) *
            (m.fs.discharge.hxd.steam_prandtl_number**(0.33)) *
            (
                (m.fs.discharge.hxd.cold_side.properties_in[0].visc_d_phase["Vap"] /
                 m.fs.discharge.hxd.cold_side.properties_out[0].visc_d_phase["Liq"]) ** 0.14
            )
        ),
        doc="Steam Nusslet Number from 2001 Zavoico, Sandia")

    # Calculate heat transfer coefficients for the salt and steam
    # sides of discharge heat exchanger
    m.fs.discharge.hxd.h_salt = pyo.Expression(
        expr=(
            m.fs.discharge.hxd.hot_side.properties_in[0].therm_cond_phase["Liq"] *
            m.fs.discharge.hxd.salt_nusselt_number /
            m.fs.discharge.hxd_tube_outer_dia
        ),
        doc="Salt side convective heat transfer coefficient in W/m.K")
    m.fs.discharge.hxd.h_steam = pyo.Expression(
        expr=(
            m.fs.discharge.hxd.cold_side.properties_in[0].therm_cond_phase["Vap"] *
            m.fs.discharge.hxd.steam_nusselt_number /
            m.fs.discharge.hxd_tube_inner_dia
        ),
        doc="Steam side convective heat transfer coefficient in W/m.K")

    # Calculate overall heat transfer coefficient for Solar salt heat
    # exchanger
    @m.fs.discharge.hxd.Constraint(m.fs.time)
    def constraint_hxd_ohtc(b, t):
        return (
            m.fs.discharge.hxd.overall_heat_transfer_coefficient[t] * (
                2 * m.fs.discharge.hxd_k_steel *
                m.fs.discharge.hxd.h_steam +
                m.fs.discharge.hxd_tube_outer_dia *
                m.fs.discharge.hxd_log_tube_dia_ratio *
                m.fs.discharge.hxd.h_salt *
                m.fs.discharge.hxd.h_steam +
                m.fs.discharge.hxd_tube_dia_ratio *
                m.fs.discharge.hxd.h_salt *
                2 * m.fs.discharge.hxd_k_steel
            )
        ) == (2 * m.fs.discharge.hxd_k_steel *
              m.fs.discharge.hxd.h_salt *
              m.fs.discharge.hxd.h_steam)


def _create_arcs(m):
    """Create arcs to connect the discharge storage system to the power
    plant

    """

    m.fs.discharge.essplit_to_hxd = Arc(
        source=m.fs.discharge.es_split.to_hxd,
        destination=m.fs.discharge.hxd.tube_inlet,
        doc="Connection from ES splitter to HXD"
    )
    m.fs.discharge.hxd_to_esturbine = Arc(
        source=m.fs.discharge.hxd.tube_outlet,
        destination=m.fs.discharge.es_turbine.inlet,
        doc="Connection from HXD to ES turbine"
    )


def disconnect_arcs(m):
    """Disconnect arcs from ultra-supercritical plant base model to
    connect the Solar salt discharge storage system

    """

    for arc_s in [
            m.fs.condpump_to_fwh1,
            m.fs.fwh4_to_fwh5,
            m.fs.booster_to_fwh6,
            m.fs.bfp_to_fwh8,
            m.fs.fwh9_to_boiler
            ]:
        arc_s.expanded_block.enth_mol_equality.deactivate()
        arc_s.expanded_block.flow_mol_equality.deactivate()
        arc_s.expanded_block.pressure_equality.deactivate()


def add_disjunction(m):
    """Add disjunction for the selection of condensate source to integrate
    the discharge storage system to the power plant model

    """

    # Add disjunction 1 for condensate source selection
    m.fs.hxd_source_disjunction = Disjunction(
        expr=[
            m.fs.discharge.booster_source_disjunct,
            m.fs.discharge.bfp_source_disjunct,
            m.fs.discharge.fwh4_source_disjunct,
            m.fs.discharge.fwh9_source_disjunct,
            m.fs.discharge.condpump_source_disjunct
            ]
    )

    # Expand arcs within the disjuncts
    expand_arcs.obj_iter_kwds['descend_into'] = (Block, Disjunct)
    TransformationFactory("network.expand_arcs").apply_to(m.fs.discharge)

    return m


def condpump_source_disjunct_equations(disj):
    """Block of equations for disjunct 1 in disjunction 1 for the selection
    of condensate water source from condenser pump

    """

    m = disj.model()

    # Define arcs to connect units within disjunct
    m.fs.discharge.condpump_source_disjunct.condpump_to_essplit = Arc(
        source=m.fs.cond_pump.outlet,
        destination=m.fs.discharge.es_split.inlet,
        doc="Connection from Condenser pump to ES splitter"
    )
    m.fs.discharge.condpump_source_disjunct.essplit_to_fwh1 = Arc(
        source=m.fs.discharge.es_split.to_fwh,
        destination=m.fs.fwh[1].tube_inlet,
        doc="Connection from ES splitter to FWH1"
    )

    m.fs.discharge.condpump_source_disjunct.fwh4_to_fwh5 = Arc(
        source=m.fs.fwh[4].tube_outlet,
        destination=m.fs.fwh[5].tube_inlet,
        doc="Connection from FWH4 to FWH5"
    )

    m.fs.discharge.condpump_source_disjunct.booster_to_fwh6 = Arc(
        source=m.fs.booster.outlet,
        destination=m.fs.fwh[6].tube_inlet,
        doc="Connection from booster pump to FWH6"
    )

    m.fs.discharge.condpump_source_disjunct.bfp_to_fwh8 = Arc(
        source=m.fs.bfp.outlet,
        destination=m.fs.fwh[8].tube_inlet,
        doc="Connection from BFP to FWH8"
    )

    m.fs.discharge.condpump_source_disjunct.fwh9_to_boiler = Arc(
        source=m.fs.fwh[9].tube_outlet,
        destination=m.fs.boiler.inlet,
        doc="Connection from FWH9 to boiler"
    )


def fwh4_source_disjunct_equations(disj):
    """Block of equations for disjunct 2 in disjunction 1 for the selection
    of condensate water source from feed water heater 4

    """

    m = disj.model()

    # Define arcs to connect units within disjunct
    m.fs.discharge.fwh4_source_disjunct.fwh4_to_essplit = Arc(
        source=m.fs.fwh[4].tube_outlet,
        destination=m.fs.discharge.es_split.inlet,
        doc="Connection from FWH4 to ES splitter"
    )
    m.fs.discharge.fwh4_source_disjunct.essplit_to_fwh5 = Arc(
        source=m.fs.discharge.es_split.to_fwh,
        destination=m.fs.fwh[5].tube_inlet,
        doc="Connection from ES splitter to FWH5"
    )

    m.fs.discharge.fwh4_source_disjunct.condpump_to_fwh1 = Arc(
        source=m.fs.cond_pump.outlet,
        destination=m.fs.fwh[1].tube_inlet,
        doc="Connection from condenser pump to FWH1"
    )

    m.fs.discharge.fwh4_source_disjunct.booster_to_fwh6 = Arc(
        source=m.fs.booster.outlet,
        destination=m.fs.fwh[6].tube_inlet,
        doc="Connection from booster pump to FWH6"
    )

    m.fs.discharge.fwh4_source_disjunct.bfp_to_fwh8 = Arc(
        source=m.fs.bfp.outlet,
        destination=m.fs.fwh[8].tube_inlet,
        doc="Connection from BFP to FWH8"
    )

    m.fs.discharge.fwh4_source_disjunct.fwh9_to_boiler = Arc(
        source=m.fs.fwh[9].tube_outlet,
        destination=m.fs.boiler.inlet,
        doc="Connection from FWH9 to boiler"
    )


def booster_source_disjunct_equations(disj):
    """Block of equations for disjunct 3 in disjunction 1 for the
    selection of condensate water source from booster pump

    """

    m = disj.model()

    # Define arcs to connect units within disjunct
    m.fs.discharge.booster_source_disjunct.booster_to_essplit = Arc(
        source=m.fs.booster.outlet,
        destination=m.fs.discharge.es_split.inlet,
        doc="Connection from Booster pump to ES splitter"
    )
    m.fs.discharge.booster_source_disjunct.essplit_to_fwh6 = Arc(
        source=m.fs.discharge.es_split.to_fwh,
        destination=m.fs.fwh[6].tube_inlet,
        doc="Connection from ES splitter to FWH6"
    )

    m.fs.discharge.booster_source_disjunct.fwh4_to_fwh5 = Arc(
        source=m.fs.fwh[4].tube_outlet,
        destination=m.fs.fwh[5].tube_inlet,
        doc="Connection from FWH4 to FWH5"
    )

    m.fs.discharge.booster_source_disjunct.condpump_to_fwh1 = Arc(
        source=m.fs.cond_pump.outlet,
        destination=m.fs.fwh[1].tube_inlet,
        doc="Connection from condenser pump to FWH1"
    )

    m.fs.discharge.booster_source_disjunct.bfp_to_fwh8 = Arc(
        source=m.fs.bfp.outlet,
        destination=m.fs.fwh[8].tube_inlet,
        doc="Connection from BFP to FWH8"
    )

    m.fs.discharge.booster_source_disjunct.fwh9_to_boiler = Arc(
        source=m.fs.fwh[9].tube_outlet,
        destination=m.fs.boiler.inlet,
        doc="Connection from FWH9 to boiler"
    )


def bfp_source_disjunct_equations(disj):
    """Block of equations for disjunct 2 in disjunction 1 for the
    selection of condensate water source from boiler feed water pump

    """

    m = disj.model()

    # Define arcs to connect units within disjunct
    m.fs.discharge.bfp_source_disjunct.bfp_to_essplit = Arc(
        source=m.fs.bfp.outlet,
        destination=m.fs.discharge.es_split.inlet,
        doc="Connection from BFP to ES splitter"
    )
    m.fs.discharge.bfp_source_disjunct.essplit_to_fwh8 = Arc(
        source=m.fs.discharge.es_split.to_fwh,
        destination=m.fs.fwh[8].tube_inlet,
        doc="Connection from ES splitter to FWH8"
    )

    m.fs.discharge.bfp_source_disjunct.fwh4_to_fwh5 = Arc(
        source=m.fs.fwh[4].tube_outlet,
        destination=m.fs.fwh[5].tube_inlet,
        doc="Connection from FWH4 to FWH5"
    )

    m.fs.discharge.bfp_source_disjunct.condpump_to_fwh1 = Arc(
        source=m.fs.cond_pump.outlet,
        destination=m.fs.fwh[1].tube_inlet,
        doc="Connection from condenser pump to FWH1"
    )

    m.fs.discharge.bfp_source_disjunct.booster_to_fwh6 = Arc(
        source=m.fs.booster.outlet,
        destination=m.fs.fwh[6].tube_inlet,
        doc="Connection from booster pump to FWH6"
    )

    m.fs.discharge.bfp_source_disjunct.fwh9_to_boiler = Arc(
        source=m.fs.fwh[9].tube_outlet,
        destination=m.fs.boiler.inlet,
        doc="Connection from FWH9 to boiler"
    )


def fwh9_source_disjunct_equations(disj):
    """Block of equations for disjunct 2 in disjunction 1 for the
    selection of condensate water source from feed water heater 9

    """

    m = disj.model()

    # Define arcs to connect units within disjunct
    m.fs.discharge.fwh9_source_disjunct.fwh9_to_essplit = Arc(
        source=m.fs.fwh[9].tube_outlet,
        destination=m.fs.discharge.es_split.inlet,
        doc="Connection from FWH9 to the ES SPlitter"
    )
    m.fs.discharge.fwh9_source_disjunct.essplit_to_boiler = Arc(
        source=m.fs.discharge.es_split.to_fwh,
        destination=m.fs.boiler.inlet,
        doc="Connection from ES splitter to Boiler"
    )

    m.fs.discharge.fwh9_source_disjunct.fwh4_to_fwh5 = Arc(
        source=m.fs.fwh[4].tube_outlet,
        destination=m.fs.fwh[5].tube_inlet,
        doc="Connection from FWH4 to FWH5"
    )

    m.fs.discharge.fwh9_source_disjunct.condpump_to_fwh1 = Arc(
        source=m.fs.cond_pump.outlet,
        destination=m.fs.fwh[1].tube_inlet,
        doc="Connection from condenser pump to FWH1"
    )

    m.fs.discharge.fwh9_source_disjunct.booster_to_fwh6 = Arc(
        source=m.fs.booster.outlet,
        destination=m.fs.fwh[6].tube_inlet,
        doc="Connection from booster to FWH6"
    )

    m.fs.discharge.fwh9_source_disjunct.bfp_to_fwh8 = Arc(
        source=m.fs.bfp.outlet,
        destination=m.fs.fwh[8].tube_inlet,
        doc="Connection from BFP to FWH8"
    )


def set_model_input(m):
    """Define model inputs such as fixed variables and parameter
    values. The arameter values in this block, unless otherwise stated
    explicitly, are either assumed or estimated for a total power out
    of 437 MW. The inputs fixed in this function are the necessary
    inputs to obtain a square model (0 degrees of freedom).

    Unless stated otherwise, the units are: temperature in K, pressure
    in Pa, flow in mol/s, massic flow in kg/s, and heat and heat duty
    in W

    """

    ###########################################################################
    # Fix data in discharge system
    ###########################################################################
    # Add heat exchanger area from supercritical plant model_input. For
    # conceptual design optimization, area is unfixed and optimized
    m.fs.discharge.hxd.area.fix(2000)

    # Define storage fluid conditions. The fluid inlet flow is fixed
    # during initialization, but is unfixed and determined during
    # optimization
    m.fs.discharge.hxd.shell_inlet.flow_mass.fix(200)
    m.fs.discharge.hxd.shell_inlet.temperature.fix(831.15)
    m.fs.discharge.hxd.shell_inlet.pressure.fix(101325)

    m.fs.discharge.es_split.inlet.flow_mol.fix(17854)
    m.fs.discharge.es_split.inlet.enth_mol.fix(52232)
    m.fs.discharge.es_split.inlet.pressure.fix(3.4958e7)

    ###########################################################################
    # Fix data in condensate source splitter
    ###########################################################################
    # The model is built for a fixed flow of condensate through the
    # discharge heat exchanger. This condensate flow is unfixed and
    # determined during design optimization
    m.fs.discharge.es_split.split_fraction[0, "to_hxd"].fix(0.2)

    ###########################################################################
    # Fix data in storage turbine
    ###########################################################################
    m.fs.discharge.es_turbine.efficiency_isentropic.fix(0.8)


def set_scaling_factors(m):
    """Scaling factors in the flowsheet

    """

    # Include scaling factors for solar discharge heat exchanger
    for htf in [m.fs.discharge.hxd]:
        iscale.set_scaling_factor(htf.area, 1e-2)
        iscale.set_scaling_factor(
            htf.overall_heat_transfer_coefficient, 1e-3)
        iscale.set_scaling_factor(htf.tube.heat, 1e-6)
        iscale.set_scaling_factor(htf.shell.heat, 1e-6)

    for est in [m.fs.discharge.es_turbine.control_volume]:
        iscale.set_scaling_factor(est.work, 1e-6)


def initialize(m, solver=None, optarg=None, outlvl=idaeslog.NOTSET):
    """Initialize the units included in the discharge model

    """

    # Include scaling factors
    iscale.calculate_scaling_factors(m)

    # Initialize splitter
    m.fs.discharge.es_split.initialize(outlvl=outlvl,
                                       optarg=optarg)

    propagate_state(m.fs.discharge.essplit_to_hxd)
    m.fs.discharge.hxd.initialize(outlvl=outlvl,
                                  optarg=optarg)

    propagate_state(m.fs.discharge.hxd_to_esturbine)
    m.fs.discharge.es_turbine.initialize(outlvl=outlvl,
                                         optarg=optarg)

    # Check and raise an error if the degrees of freedom are not 0
    if not degrees_of_freedom(m) == 0:
        raise ConfigurationError(
            "The degrees of freedom after building the model are not 0. "
            "You have {} degrees of freedom. "
            "Please check your inputs to ensure a square problem "
            "before initializing the model.".format(degrees_of_freedom(m))
            )

    # Solve initialization
    init_results = solver.solve(m, options=optarg)
    print("Discharge model initialization solver termination:",
          init_results.solver.termination_condition)
    print("*************   Discharge Model Initialized   ******************")


def build_costing(m, solver=None):
    """Add cost correlations for the storage design analysis

    This function is used to estimate the capital and operating cost
    of integrating a discharge storage system to the power plant and
    it contains cost correlations to estimate: (i) the capital cost of
    discharge heat exchanger and Solar salt pump, and (ii) the
    operating costs for 1 year

    """

    ###########################################################################
    # Add capital cost
    # 1. Calculate discharge heat exchanger cost
    # 2. Calculate Solar salt pump purchase cost
    # 3. Calculate total capital cost of discharge system
    
    # Main assumptions
    # 1. Salt life is assumed to outlast the plant life
    # 2. The economic objective is to minimize total annualized cost. So, cash
    # flows, discount rate, and NPV are not included in this study.
    ###########################################################################
    # Add capital cost: 1. Calculate discharge heat exchanger cost
    ###########################################################################
    # Calculate and initialize Solar salt discharge heat exchanger
    # cost, which is estimated using the IDAES costing method with
    # default options, i.e. a U-tube heat exchanger, stainless steel
    # material, and a tube length of 12ft. Refer to costing
    # documentation to change any of the default options. The purchase
    # cost of heat exchanger has to be annualized when used
    m.fs.costing = SSLWCosting()

    m.fs.discharge.hxd.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=SSLWCostingData.cost_heat_exchanger,
    )

    ###########################################################################
    # Add capital cost: 2. Calculate Solar salt pump purchase cost
    ###########################################################################
    # Pump for moving Solar salt is not explicity modeled. To compute
    # the capital costs for this pump the capital cost expressions are
    # added below.  All cost expressions are from the same reference
    # as the IDAES costing framework and is given below: Seider,
    # Seader, Lewin, Windagdo, 3rd Ed. John Wiley and Sons, Chapter
    # 22. Cost Accounting and Capital Cost Estimation, Section 22.2 Cost
    # Indexes and Capital Investment

    # ---------- Solar salt ----------
    # Calculate purchase cost of Solar salt pump
    m.fs.discharge.spump_Qgpm = pyo.Expression(
        expr=(m.fs.discharge.hxd.
              hot_side.properties_in[0].flow_mass *
              (264.17 * pyo.units.gallon / pyo.units.m**3) *
              (60 * pyo.units.s / pyo.units.min) /
              (m.fs.discharge.hxd.
               hot_side.properties_in[0].dens_mass["Liq"])),
        doc="Conversion of Solar salt flow mass to volumetric flow in gallons/min"
    )
    m.fs.discharge.dens_lbft3 = pyo.units.convert(
        m.fs.discharge.hxd.hot_side.properties_in[0].dens_mass["Liq"],
        to_units=pyo.units.pound / pyo.units.foot**3
    )
    m.fs.discharge.spump_sf = pyo.Expression(
        expr=(m.fs.discharge.spump_Qgpm *
              (m.fs.discharge.spump_head ** 0.5)),
        doc="Pump size factor"
    )
    # Expression for pump base purchase cost
    m.fs.discharge.pump_CP = pyo.Expression(
        expr=(
            m.fs.discharge.spump_FT * m.fs.discharge.spump_FM *
            exp(
                9.7171 -
                0.6019 * log(m.fs.discharge.spump_sf) +
                0.0519 * ((log(m.fs.discharge.spump_sf))**2)
            )
        ),
        doc="Base purchase cost of Solar salt pump in $"
    )
    # Expression for pump efficiency
    m.fs.discharge.spump_np = pyo.Expression(
        expr=(
            -0.316 +
            0.24015 * log(m.fs.discharge.spump_Qgpm) -
            0.01199 * ((log(m.fs.discharge.spump_Qgpm))**2)
        ),
        doc="Fractional efficiency of the pump in horsepower"
    )
    m.fs.discharge.motor_pc = pyo.Expression(
        expr=(
            (m.fs.discharge.spump_Qgpm *
             m.fs.discharge.spump_head *
             m.fs.discharge.dens_lbft3) /
            (33000 *
             m.fs.discharge.spump_np *
             m.fs.discharge.spump_nm)
        ),
        doc="Power consumption of motor in horsepower"
    )

    # Defining a local variable for the log of motor's power consumption
    # This will help writing the motor's purchase cost expressions conciesly
    _log_motor_pc = log(m.fs.discharge.motor_pc)

    # Expression for motor's purchase cost
    m.fs.discharge.motor_CP = pyo.Expression(
        expr=(
            m.fs.discharge.spump_motorFT *
            exp(
                5.8259 +
                0.13141 * _log_motor_pc +
                0.053255 * (_log_motor_pc**2) +
                0.028628 * (_log_motor_pc**3) -
                0.0035549 * (_log_motor_pc**4)
            )
        ),
        doc="Base cost of Solar salt pump's motor in $"
    )

    # Calculate and initialize total cost of Solar salt pump
    m.fs.discharge.spump_purchase_cost = pyo.Var(
        initialize=100000,
        bounds=(0, 1e7),
        doc="Total purchase cost of Solar salt pump in $"
    )

    def solar_spump_purchase_cost_rule(b):
        return (
            m.fs.discharge.spump_purchase_cost == (
                m.fs.discharge.pump_CP +
                m.fs.discharge.motor_CP) *
            (m.CE_index / 394)
        )
    m.fs.discharge.spump_purchase_cost_eq = pyo.Constraint(
        rule=solar_spump_purchase_cost_rule)

    calculate_variable_from_constraint(
        m.fs.discharge.spump_purchase_cost,
        m.fs.discharge.spump_purchase_cost_eq)

    ###########################################################################
    # Add capital cost: 3. Calculate total capital cost for discharge system
    ###########################################################################

    # Add capital cost variable at flowsheet level to handle the Solar
    # salt capital cost
    m.fs.discharge.capital_cost = pyo.Var(
        initialize=1000000,
        bounds=(0, 1e10),
        doc="Annualized capital cost in $/year")

    # Calculate and initialize annualized capital cost for the Solar
    # salt discharge storage system
    def solar_cap_cost_rule(b):
        return (
            m.fs.discharge.capital_cost *
            m.fs.discharge.num_of_years
        ) == (m.fs.discharge.spump_purchase_cost +
              m.fs.discharge.hxd.costing.capital_cost)
    m.fs.discharge.cap_cost_eq = pyo.Constraint(
        rule=solar_cap_cost_rule)

    calculate_variable_from_constraint(
        m.fs.discharge.capital_cost,
        m.fs.discharge.cap_cost_eq)

    ###########################################################################
    #  Add operating cost
    ###########################################################################
    m.fs.discharge.operating_hours = pyo.Expression(
        expr=365 * 3600 * m.fs.discharge.hours_per_day,
        doc="Number of operating hours per year")
    m.fs.discharge.operating_cost = pyo.Var(
        initialize=1000000,
        bounds=(0, 1e11),
        doc="Operating cost in $/year")

    def op_cost_rule(b):
        return m.fs.discharge.operating_cost == (
            m.fs.discharge.operating_hours *
            m.fs.discharge.coal_price *
            m.fs.coal_heat_duty * 1e6
        )
    m.fs.discharge.op_cost_eq = pyo.Constraint(rule=op_cost_rule)

    # Initialize operating cost
    calculate_variable_from_constraint(
        m.fs.discharge.operating_cost,
        m.fs.discharge.op_cost_eq)

    # Check and raise an error if the degrees of freedom are not 0
    if not degrees_of_freedom(m) == 0:
        raise ConfigurationError(
            "The degrees of freedom after building costing block are not 0. "
            "You have {} degrees of freedom. "
            "Please check your inputs to ensure a square problem "
            "before initializing the model.".format(degrees_of_freedom(m))
            )

    # Solve cost initialization
    print()
        # Add options to NLP solver
    optarg = {"tol": 1e-8,
              "max_iter": 300}
    cost_results = solver.solve(m, options=optarg)
    print("Cost initialization solver termination:",
          cost_results.solver.termination_condition)
    print("******************** Costing Initialized *************************")
    print()
    print()


def add_bounds(m, power_max=None):
    """Add bounds to all units in discharge model

    """

    m.flow_max = m.main_flow * 2        # Units in mol/s
    m.storage_flow_max = 0.2 * m.flow_max # Units in mol/s
    m.salt_flow_max = 1000                # Units in kg/s
    m.heat_duty_bound = 200e6             # Units in MW
    m.power_max = power_max               # Units in MW

    # Add bounds to Solar salt discharge heat exchanger
    for hxd in [m.fs.discharge.hxd]:
        hxd.tube_inlet.flow_mol.setlb(0)
        hxd.tube_inlet.flow_mol.setub(m.storage_flow_max)
        hxd.shell_inlet.flow_mass.setlb(0)
        hxd.shell_inlet.flow_mass.setub(m.salt_flow_max)
        hxd.tube_outlet.flow_mol.setlb(0)
        hxd.tube_outlet.flow_mol.setub(m.storage_flow_max)
        hxd.shell_outlet.flow_mass.setlb(0)
        hxd.shell_outlet.flow_mass.setub(m.salt_flow_max)
        hxd.shell_inlet.pressure.setlb(101320)
        hxd.shell_inlet.pressure.setub(101330)
        hxd.shell_outlet.pressure.setlb(101320)
        hxd.shell_outlet.pressure.setub(101330)
        hxd.heat_duty.setlb(0)
        hxd.heat_duty.setub(m.heat_duty_bound)
        hxd.shell.heat.setlb(-m.heat_duty_bound)
        hxd.shell.heat.setub(0)
        hxd.tube.heat.setlb(0)
        hxd.tube.heat.setub(m.heat_duty_bound)
        hxd.shell.properties_in[0].enth_mass.setlb(0)
        hxd.shell.properties_in[0].enth_mass.setub(1.5e6)
        hxd.shell.properties_out[0].enth_mass.setlb(0)
        hxd.shell.properties_out[0].enth_mass.setub(1.5e6)
        hxd.overall_heat_transfer_coefficient.setlb(0)
        hxd.overall_heat_transfer_coefficient.setub(10000)
        hxd.area.setlb(0)
        hxd.area.setub(5000)
        hxd.costing.pressure_factor.setlb(0)
        hxd.costing.pressure_factor.setub(1e5)
        hxd.costing.capital_cost.setlb(0)
        hxd.costing.capital_cost.setub(1e7)
        hxd.costing.base_cost_per_unit.setlb(0)
        hxd.costing.base_cost_per_unit.setub(1e6)
        hxd.costing.material_factor.setlb(0)
        hxd.costing.material_factor.setub(10)
        hxd.delta_temperature_in.setlb(10)
        hxd.delta_temperature_out.setlb(9)
        hxd.delta_temperature_in.setub(298)
        hxd.delta_temperature_out.setub(500)

    # Add bounds needed in units declared in condensate source
    # disjunction
    for split in [m.fs.discharge.es_split]:
        split.inlet.flow_mol[:].setlb(0)
        split.inlet.flow_mol[:].setub(m.flow_max)
        split.to_hxd.flow_mol[:].setlb(0)
        split.to_hxd.flow_mol[:].setub(m.storage_flow_max)
        split.to_fwh.flow_mol[:].setlb(0)
        split.to_fwh.flow_mol[:].setub(m.flow_max)
        split.split_fraction[0.0, "to_hxd"].setlb(0)
        split.split_fraction[0.0, "to_hxd"].setub(1)
        split.split_fraction[0.0, "to_fwh"].setlb(0)
        split.split_fraction[0.0, "to_fwh"].setub(1)

    m.fs.plant_power_out[0].setlb(300)
    m.fs.plant_power_out[0].setub(m.power_max)

    for unit_k in [m.fs.booster]:
        unit_k.inlet.flow_mol[:].setlb(0)
        unit_k.inlet.flow_mol[:].setub(m.flow_max)
        unit_k.outlet.flow_mol[:].setlb(0)
        unit_k.outlet.flow_mol[:].setub(m.flow_max)

    for k in m.set_turbine:
        m.fs.turbine[k].work.setlb(-1e10)
        m.fs.turbine[k].work.setub(0)


def main(m_usc, solver=None, optarg=None):

    # Add boiler and cycle efficiencies to the model
    add_efficiency = True

    # Add maximum power produced by power plant in MW. For this
    # analysis, the maximum power is fixed to 436 MW
    power_max = 436

    # Create a flowsheet, add properties, unit models, and arcs
    m = create_discharge_model(m_usc,
                               add_efficiency=add_efficiency,
                               power_max=power_max)

    # Give all the required inputs to the model
    set_model_input(m)

    # Add scaling factor
    set_scaling_factors(m)

    # Initialize the model with a sequential initialization
    initialize(m, solver=solver, optarg=optarg)

    # Add cost correlations
    build_costing(m, solver=solver)

    # Add bounds
    add_bounds(m, power_max=power_max)

    # Disconnect arcs to include discharge storage system
    disconnect_arcs(m)

    # Add disjunction
    add_disjunction(m)

    return m


def print_model(_, nlp_model, nlp_data):
    """Print the disjunction selected during the solution of the NLP
    subproblem

    """

    nlp = nlp_model.fs.discharge
    print('    ___________________________________________')
    print('     Disjunction 1:')
    if nlp.condpump_source_disjunct.binary_indicator_var.value == 1:
        print('      Condensate from condenser pump is selected')
    elif nlp.booster_source_disjunct.binary_indicator_var.value == 1:
        print('      Condensate from booster pump is selected')
    elif nlp.bfp_source_disjunct.binary_indicator_var.value == 1:
        print('      Condensate from boiler feed pump is selected')
    elif nlp.fwh9_source_disjunct.binary_indicator_var.value == 1:
        print('      Condensate from FWH9 is selected')
    elif nlp.fwh4_source_disjunct.binary_indicator_var.value == 1:
        print('      Condensate from FWH4 is selected')
    else:
        print('      Error: There are no more alternatives')
    print('    ___________________________________________')
    print()


def run_gdp(m):
    """Declare solver GDPopt and its options
    """

    # Add options to GDPopt
    opt = SolverFactory('gdpopt')
    _prop_bnds_root_to_leaf_map[ExternalFunctionExpression] = lambda x, y, z: None

    # Solve model
    results = opt.solve(
        m,
        tee=True,
        algorithm='RIC',
        init_algorithm="no_init",
        mip_solver='cbc',
        nlp_solver='ipopt',
        call_after_subproblem_solve=print_model,
        subproblem_presolve = False,
        nlp_solver_args=dict(
            tee=True,
            options={
                "max_iter": 150}
        )
    )

    return results


def print_results(m, results):

    print('====================================================')
    print('Results ')
    print()
    print('Obj (M$/year): {:.2f}'.format(
        (pyo.value(m.obj) / scaling_obj) * 1e-6))
    print('Discharge capital cost (M$/y): {:.2f}'.format(
        pyo.value(m.fs.discharge.capital_cost) * 1e-6))
    print('Net Power (MW): {:.2f}'.format(
        pyo.value(m.fs.net_power)))
    print('Plant Power (MW): {:.2f}'.format(
        pyo.value(m.fs.plant_power_out[0])))
    print('Discharge Turbine Power (MW): {:.2f}'.format(
        pyo.value(m.fs.discharge.es_turbine.control_volume.work[0]) * (-1e-6)))
    print('Boiler Efficiency (%): {:.2f}'.format(
        pyo.value(m.fs.boiler_efficiency) * 100))
    print()
    print("**Discrete design decisions (Disjunction)")
    for d in m.component_data_objects(ctype=Disjunct,
                                      active=True,
                                      sort=True, descend_into=True):
        if abs(d.binary_indicator_var.value - 1) < 1e-6:
            print(d.name, ' should be selected!')
    print('Discharge heat exchanger area (m2): {:.2f}'.format(
        pyo.value(m.fs.discharge.hxd.area)))
    print('Discharge heat exchanger heat duty (MW): {:.2f}'.format(
        pyo.value(m.fs.discharge.hxd.heat_duty[0]) * 1e-6))
    print('====================================================')
    print()
    print('Solver details')
    print(results)
    print()


def model_analysis(m, heat_duty=None):
    """Solve the conceptual design optimization problem

    """

    # Fix variables in the flowsheet
    m.fs.plant_power_out.fix(400)
    m.fs.boiler.outlet.pressure.fix(m.main_steam_pressure)
    m.fs.discharge.hxd.heat_duty.fix(heat_duty * 1e6)

    # Unfix variables that were fixed iduring initialization
    m.fs.boiler.inlet.flow_mol.unfix()
    m.fs.discharge.es_split.split_fraction[0, "to_hxd"].unfix()
    m.fs.discharge.es_split.inlet.unfix()
    m.fs.discharge.hxd.shell_inlet.flow_mass.unfix()
    m.fs.discharge.hxd.area.unfix()

    # Add total cost as the objective function
    m.obj = Objective(
        expr=(
            m.fs.discharge.capital_cost +
            m.fs.discharge.operating_cost
        ) * scaling_obj
    )


if __name__ == "__main__":

    # optarg = {"max_iter": 300}
    optarg = {"tol": 1e-8,
              "max_iter": 300,
              "halt_on_ampl_error": "yes"}
    solver = get_solver('ipopt', optarg)

    # Build ultra-supercritical plant base model
    m_usc = usc.build_plant_model()

    # Initialize ultra-supercritical plant base model
    usc.initialize(m_usc)

    # Build discharge model
    m = main(m_usc, solver=solver, optarg=optarg)

    # Solve design model optimization problem
    heat_duty_data = 148.5
    model_analysis(m, heat_duty=heat_duty_data)

    # Solve model using GDPopt
    print()
    print('**********Start solution of GDP discharge model using GDPopt')
    print('DOFs before GDP discharge model solution: ', degrees_of_freedom(m))
    print()
    results = run_gdp(m)

    # Print results
    print_results(m, results)



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################

"""
Thermo-physical property package for Hitec Salt based on empirical correlations
obtained from literatue. Hitec salt is a tertiary salt mixture consisting
of (% wt.): 40% NaNO2, 7% NaNO3, and 53% KNO3
References:
(1) Sohal et al., (2010) Engineering Database of Liquid Salt Thermophysical
and Thermochemical Properties
(2) Change et al., (2015) Energy Procedia, 69, 779 - 789
updated on 05/18/2021
"""


# Import Python libraries
import logging
# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           Param,
                           PositiveReals,
                           Expression,
                           Var,
                           exp,
                           log,
                           units as pyunits,
                           Reals)
from pyomo.opt import TerminationCondition

# Import IDAES
from idaes.core import (declare_process_block_class,
                        StateBlock,
                        MaterialBalanceType,
                        EnergyBalanceType,
                        StateBlockData,
                        PhysicalParameterBlock)
from idaes.core.util.initialization import solve_indexed_blocks
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import fix_state_vars, revert_state_vars
from idaes.core.base.phases import LiquidPhase
from idaes.core.base.components import Component
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog

# Some more inforation about this module
_author_ = "Naresh Susarla"
__version__ = "0.0.1"


# Logger
_log = logging.getLogger(__name__)
init_log = idaeslog.getInitLogger(__name__)
# *****************************************************************************


@declare_process_block_class("HitecsaltParameterBlock")
class PhysicalParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class.
    Contains parameters associated with properties for Hitec Salt.
    """

    def build(self):
        """Callable method for Block construction."""
        super(PhysicalParameterData, self).build()
        self._state_block_class = HitecsaltStateBlock
        self._make_params()
        self.set_default_scaling('flow_mass', 0.1)
        self.set_default_scaling('temperature', 0.01)
        self.set_default_scaling('pressure', 1e-5)
        self.set_default_scaling('enth_mass', 1e-5)
        self.set_default_scaling('dens_mass', 1e-3)
        self.set_default_scaling('cp_mass', 1e-3)
        self.set_default_scaling('visc_d_phase', 10)
        self.set_default_scaling('therm_cond_phase', 10)
        self.set_default_scaling('enthalpy_flow_terms', 1e-6)

    def _make_params(self):

        # Add Phase objects
        self.Liq = LiquidPhase()
        self.Hitec_Salt = Component()

#        Specific heat capacity at constant pressure (cp) coefficients
#        Cp in J/kg/K
        self.cp_param_1 = Param(initialize=5806,
                                units=pyunits.J/(pyunits.kg*pyunits.K),
                                doc="Coefficient: specific heat expression")
        self.cp_param_2 = Param(initialize=-10.833,
                                units=pyunits.J/(pyunits.kg*(pyunits.K**2)),
                                doc="Coefficient: specific heat expression")
        self.cp_param_3 = Param(initialize=7.2413E-3,
                                units=pyunits.J/(pyunits.kg*(pyunits.K**3)),
                                doc="Coefficient: specific heat expression")

#        Density (rho) coefficients
#        rho in kg/m3
        self.rho_param_1 = Param(initialize=2293.6,
                                 units=pyunits.kg/(pyunits.m**3),
                                 doc="Coefficient: density expression")
        self.rho_param_2 = Param(initialize=-0.7497,
                                 units=pyunits.kg/(pyunits.K*(pyunits.m**3)),
                                 doc="Coefficient: density expression")

#        Dynamic Viscosity (mu) coefficients
#        Mu in Pa.s
        self.mu_param_1 = Param(initialize=-4.343,
                                units=pyunits.Pa*pyunits.s,
                                doc="Coefficient: dynamic viscosity")
        self.mu_param_2 = Param(initialize=-2.0143,
                                units=pyunits.Pa*pyunits.s/pyunits.K,
                                doc="Coefficient: dynamic viscosity")
        self.mu_param_3 = Param(initialize=-5.011,
                                units=pyunits.Pa*pyunits.s/(pyunits.K**2),
                                doc="Coefficient: dynamic viscosity")

#        Thermal conductivity (kappa) coefficients
#        kappa in W/(m.K)
        self.kappa_param_1 = Param(initialize=0.421,
                                   units=pyunits.W/(pyunits.m*pyunits.K),
                                   doc="Coefficient: thermal conductivity")
        self.kappa_param_2 = Param(initialize=-6.53E-4,
                                   units=pyunits.W/(pyunits.m*(pyunits.K**2)),
                                   doc="Coefficient: thermal conductivity")
        self.kappa_param_3 = Param(initialize=-260,
                                   units=pyunits.W/(pyunits.m*(pyunits.K**3)),
                                   doc="Coefficient: thermal conductivity")

#        Thermodynamic reference state
#        ref_temperature in K
        self.ref_temperature = Param(within=PositiveReals,
                                     mutable=True,
                                     default=298.15,
                                     units=pyunits.K,
                                     doc='Reference temperature')

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'flow_mass': {'method': None, 'units': 'kg/s'},
                            'temperature': {'method': None, 'units': 'K'},
                            'pressure': {'method': None, 'units': 'Pa'},
                            'cp_mass': {'method': None,
                                                 'units': 'J/kg/K'},
                            'dens_mass': {'method': None, 'units': 'kg/m3'},
                            'enth_mass': {'method': None, 'units': 'J/kg'},
                            'visc_d_phase': {'method': None,
                                             'units': 'Pa.s'},
                            'therm_cond_phase': {'method': None,
                                                 'units': 'W/m/K'}})
        obj.add_default_units({'time': pyunits.s,
                               'length': pyunits.m,
                               'mass': pyunits.kg,
                               'amount': pyunits.mol,
                               'temperature': pyunits.K})


class _StateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """
    def initialize(blk, state_args=None,
                   outlvl=idaeslog.WARNING,
                   hold_state=False,
                   state_vars_fixed=False,
                   solver=None,
                   optarg=None):
        """
        Declare initialisation routine.
        Keyword Arguments:
            state_args = to be used if state block initialized independent of
                         control volume initialize
            outlvl : sets output level of initialisation routine
                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)
            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        # Fix state variables if not already fixed by the control volume block
        if state_vars_fixed is False:
            # Fix state variables if not already fixed
            flags = fix_state_vars(blk, state_args)
        for k in blk.keys():
            if degrees_of_freedom(blk[k]) != 0:
                raise Exception("State vars fixed but degrees of freedom "
                                "for state block is not zero during "
                                "initialization.")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Solve property correlation
        results = solve_indexed_blocks(opt, [blk])
        if (results.solver.termination_condition ==
                TerminationCondition.optimal):
            init_log.info('{} Initialisation Step 1 Complete.'.
                          format(blk.name))
        else:
            init_log.warning('{} Initialisation Step 1 Failed.'.
                             format(blk.name))

        init_log.info('Initialization Step 1 Complete.')

        # ---------------------------------------------------------------------
        if state_vars_fixed is False:
            # release state vars fixed during initialization if control
            # volume didn't fix the state vars
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)
        init_log.info('Initialization Complete.')

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        # Method to release states only if explicitly called

        if flags is None:
            return
        # Unfix state variables
        revert_state_vars(blk, flags)
        init_log.info('State Released.')


@declare_process_block_class("HitecsaltStateBlock", block_class=_StateBlock)
class HitecsaltStateBlockData(StateBlockData):
    """An example property package for Hitec Salt properties."""

    def build(self):
        """Callable method for Block construction."""
        super(HitecsaltStateBlockData, self).build()
        self._make_state_vars()
        self._make_prop_vars()
        self._make_constraints()

    def _make_state_vars(self):
        """Declare the necessary state variable objects."""
        self.flow_mass = Var(domain=Reals,
                             initialize=100,
                             units=pyunits.kg/pyunits.s,
                             doc='Fluid mass flowrate')
        self.pressure = Var(domain=Reals,
                            initialize=1.01325E5,
                            units=pyunits.Pa,
                            doc='State pressure')
        self.temperature = Var(domain=Reals,
                               initialize=550,
                               units=pyunits.K,
                               doc='State temperature',
                               bounds=(435.15, 788.15))  # i.e., 162 - 515 C

    def _make_prop_vars(self):
        """Make additional variables for calcuations."""

        self.enth_mass = Var(self.phase_list,
                                 initialize=1,
                                 units=pyunits.J/pyunits.kg,
                                 doc='Specific Enthalpy')

    def _make_constraints(self):
        """Create property constraints."""

        # Specific heat capacity
        self.cp_mass = Expression(
            self.phase_list,
            expr=(self.params.cp_param_1 +
                  (self.params.cp_param_2 * self.temperature) +
                  (self.params.cp_param_3 * (self.temperature**2))),
            doc="Specific heat capacity")

        # Density
        self.dens_mass = Expression(
            self.phase_list,
            expr=(self.params.rho_param_1 +
                  (self.params.rho_param_2 * self.temperature)),
            doc="density")

        # Specific Enthalpy
        def enthalpy_correlation(self, p):
            return (
                self.enth_mass[p] ==
                ((self.params.cp_param_1 * self.temperature) +
                 (self.params.cp_param_2 * self.temperature**2) +
                 (self.params.cp_param_3 * self.temperature**3)))
        self.enthalpy_eq = Constraint(self.phase_list,
                                      rule=enthalpy_correlation)

#           Dynamic viscosity and thermal conductivity
#           Ref: (2015) Change et al., Energy Procedia, 69, 779 - 789

        # Dynamic viscosity
        self.visc_d_phase = Expression(
            self.phase_list,
            expr=exp(self.params.mu_param_1 +
                     self.params.mu_param_2 * (log(self.temperature) +
                                               self.params.mu_param_3)),
            doc="dynamic viscosity")

        # Thermal conductivity
        self.therm_cond_phase = Expression(
            self.phase_list,
            expr=self.params.kappa_param_1 +
            self.params.kappa_param_2 * (self.temperature +
                                         self.params.kappa_param_3),
            doc="dynamic viscosity")

        # Enthalpy flow terms
        def rule_enthalpy_flow_terms(b, p):
            return (self.enth_mass[p] * self.flow_mass)

        self.enthalpy_flow_terms = Expression(
            self.config.parameters.phase_list,
            rule=rule_enthalpy_flow_terms)

    def get_material_flow_terms(b, p, j):
        """Define material flow terms for control volume."""
        return b.flow_mass

    def get_enthalpy_flow_terms(b, p):
        """Define enthalpy flow terms for control volume."""
        return b.enthalpy_flow_terms[p]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(b):
        """Define state variables for ports."""
        return {"flow_mass": b.flow_mass,
                "temperature": b.temperature,
                "pressure": b.pressure}



#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
# Import objects from pyomo package
from pytest import approx
from pyomo.environ import ConcreteModel, SolverFactory, Var, TerminationCondition, SolverStatus, value

# Import the main FlowsheetBlock from IDAES. The flowsheet block will contain the unit model
from idaes.core import FlowsheetBlock
from idaes.core.util.testing import initialization_tester

from dispatches.unit_models import ElectricalSplitter


def test_elec_splitter_num_outlets_build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)  # dynamic or ss flowsheet needs to be specified here

    m.fs.unit = ElectricalSplitter(num_outlets=3)

    assert hasattr(m.fs.unit, "electricity")
    assert hasattr(m.fs.unit, "electricity_in")
    assert hasattr(m.fs.unit, "outlet_list")
    assert isinstance(m.fs.unit.electricity, Var)
    assert isinstance(m.fs.unit.outlet_1_elec, Var)
    assert isinstance(m.fs.unit.outlet_2_elec, Var)
    assert isinstance(m.fs.unit.outlet_3_elec, Var)


def test_elec_splitter_num_outlets_init_0():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.unit = ElectricalSplitter(num_outlets=3)

    # fix 2 outlets, dof=0
    m.fs.unit.electricity_in.electricity.fix(1)
    m.fs.unit.outlet_1_elec.fix(0.25)
    m.fs.unit.outlet_2_elec.fix(0.25)
    initialization_tester(m)


def test_elec_splitter_num_outlets_init_1():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.unit = ElectricalSplitter(num_outlets=3)

    # fix 1 outlets, dof=1
    m.fs.unit.electricity_in.electricity.fix(1)
    m.fs.unit.outlet_1_elec.fix(0.25)
    initialization_tester(m, dof=1)


def test_elec_splitter_num_outlets_init_3():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.unit = ElectricalSplitter(num_outlets=3, add_split_fraction_vars=True)

    # fix 2 outlets, dof=0
    m.fs.unit.electricity_in.electricity.fix(1)
    m.fs.unit.split_fraction['outlet_1', 0].fix(0.25)
    m.fs.unit.split_fraction['outlet_2', 0].fix(0.25)
    initialization_tester(m)


def test_elec_splitter_num_outlets_init_4():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.unit = ElectricalSplitter(num_outlets=3, add_split_fraction_vars=True)

    # fix 1 outlets, dof=1
    m.fs.unit.electricity_in.electricity.fix(1)
    m.fs.unit.split_fraction['outlet_1', 0].fix(0.25)
    initialization_tester(m, dof=1)


def test_elec_splitter_num_outlets_solve_0():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.unit = ElectricalSplitter(num_outlets=3)

    m.fs.unit.electricity_in.electricity.fix(1)
    m.fs.unit.outlet_1_elec.fix(0.25)
    m.fs.unit.outlet_2_elec.fix(0.25)

    solver = SolverFactory('ipopt')
    results = solver.solve(m.fs)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert m.fs.unit.outlet_1_elec[0].value == 0.25
    assert m.fs.unit.outlet_2_elec[0].value == 0.25
    assert m.fs.unit.outlet_3_elec[0].value == approx(0.5, 1e-4)


def test_elec_splitter_num_outlets_solve_1():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.unit = ElectricalSplitter(num_outlets=3)

    m.fs.unit.outlet_1_elec.fix(25)
    m.fs.unit.outlet_2_elec.fix(25)
    m.fs.unit.outlet_3_elec.fix(50)

    solver = SolverFactory('ipopt')
    results = solver.solve(m.fs)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert m.fs.unit.electricity[0].value == 100


def test_elec_splitter_outlet_list():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)  # dynamic or ss flowsheet needs to be specified here
    m.fs.unit = ElectricalSplitter(outlet_list=["o1", "o2"])

    assert hasattr(m.fs.unit, "electricity")
    assert hasattr(m.fs.unit, "electricity_in")
    assert hasattr(m.fs.unit, "split_fraction")
    assert hasattr(m.fs.unit, "outlet_list")
    assert isinstance(m.fs.unit.electricity, Var)
    assert isinstance(m.fs.unit.o1_elec, Var)
    assert isinstance(m.fs.unit.o2_elec, Var)

    m.fs.unit.o1_elec.fix(0.5)
    m.fs.unit.o2_elec.fix(0.5)

    solver = SolverFactory('ipopt')
    results = solver.solve(m.fs)

    assert results.solver.termination_condition == TerminationCondition.optimal
    assert results.solver.status == SolverStatus.ok

    assert m.fs.unit.electricity[0].value == approx(1, 1e-4)
    assert value(m.fs.unit.split_fraction['o1', 0]) == approx(0.5, 1e-4)
    assert value(m.fs.unit.split_fraction['o2', 0]) == approx(0.5, 1e-4)

    m.fs.unit.report(dof=True)




#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
# expose selected objects in package namespace
from .workflow import ManagedWorkflow


#################################################################################
# DISPATCHES was produced under the DOE Design Integration and Synthesis
# Platform to Advance Tightly Coupled Hybrid Energy Systems program (DISPATCHES),
# and is copyright (c) 2022 by the software owners: The Regents of the University
# of California, through Lawrence Berkeley National Laboratory, National
# Technology & Engineering Solutions of Sandia, LLC, Alliance for Sustainable
# Energy, LLC, Battelle Energy Alliance, LLC, University of Notre Dame du Lac, et
# al. All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. Both files are also available online at the URL:
# "https://github.com/gmlc-dispatches/dispatches".
#
#################################################################################
"""
Property package for Therminol-66
Authored by: Konor Frick and Jaffer Ghouse
Edited by: Naresh Susarla and Soraya Rawlings
Date: 12/02/2021
Source:
    1. Therminol 66, High Performance Highly Stable Heat Transfer Fluid
    (0C to 345C), Solutia.
"""

import logging
# Import Pyomo libraries
from pyomo.environ import (Constraint,
                           Param,
                           PositiveReals,
                           Expression,
                           Reals,
                           units as pyunits,
                           value,
                           Var,
                           exp)
from pyomo.opt import TerminationCondition
# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        PhysicalParameterBlock,
                        StateBlockData,
                        StateBlock,
                        MaterialBalanceType,
                        EnergyBalanceType)
from idaes.core.base.phases import LiquidPhase
from idaes.core.base.components import Component
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.initialization import solve_indexed_blocks, \
    fix_state_vars, revert_state_vars
from idaes.core.solvers import get_solver
import idaes.logger as idaeslog

# Some more information about this module
__author__ = "Naresh Susarla, Soraya Rawlings, Jaffer Ghouse, Konor Frick"


# Logger
_log = logging.getLogger(__name__)
init_log = idaeslog.getInitLogger(__name__)


@declare_process_block_class("ThermalOilParameterBlock")
class PhysicalParameterData(PhysicalParameterBlock):
    """
    Property Parameter Block Class

    """

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(PhysicalParameterData, self).build()
        self._state_block_class = ThermalOilStateBlock
        # Add Phase objects
        self.Liq = LiquidPhase()
        # Add Component objects
        self.therminol66 = Component()
        # Add scaling factors
        self._make_params()
        self.set_default_scaling('flow_mass', 0.1)
        self.set_default_scaling('temperature', 0.01)
        self.set_default_scaling('pressure', 1e-5)
        self.set_default_scaling('enth_mass', 1e-5)
        self.set_default_scaling('dens_mass', 1e-3)
        self.set_default_scaling('cp_mass', 1e-3)
        self.set_default_scaling('visc_k_phase', 10)
        self.set_default_scaling('visc_d_phase', 10)
        self.set_default_scaling('therm_cond_phase', 10)
        self.set_default_scaling('enthalpy_flow_terms', 1e-6)

    def _make_params(self):
#        Specific heat capacity at constant pressure (cp) coefficients
#        Cp in J/kg/K
        self.cp_param_1 = Param(initialize=1496.005,
                                units=pyunits.J/(pyunits.kg*pyunits.K),
                                doc="Coefficient: specific heat expression")
        self.cp_param_2 = Param(initialize=3.313,
                                units=pyunits.J/(pyunits.kg*(pyunits.K**2)),
                                doc="Coefficient: specific heat expression")
        self.cp_param_3 = Param(initialize=0.0008970785,
                                units=pyunits.J/(pyunits.kg*(pyunits.K**3)),
                                doc="Coefficient: specific heat expression")

#        Density (rho) coefficients
#        rho in kg/m3
        self.rho_param_1 = Param(initialize=1026.7,
                                 units=pyunits.kg/(pyunits.m**3),
                                 doc="Coefficient: density expression")
        self.rho_param_2 = Param(initialize=-0.7281,
                                 units=pyunits.kg/(pyunits.K*(pyunits.m**3)),
                                 doc="Coefficient: density expression")

#        Kinematic viscosity (nu) coefficients
#        Nu in m2/s
        self.nu_param_1 = Param(initialize=586.375,
                                units=pyunits.K,
                                doc="Coefficient: kinematic viscosity")
        self.nu_param_2 = Param(initialize=62.5,
                                units=pyunits.K,
                                doc="Coefficient: kinematic viscosity")
        self.nu_param_3 = Param(initialize=-2.2809,
                                units=pyunits.K/pyunits.K,
                                doc="Coefficient: kinematic viscosity")
        self.nu_param_4 = Param(initialize=1E-6,
                                units=(pyunits.m**2)/pyunits.s,
                                doc="Coefficient: kinematic viscosity")

#        Thermal conductivity (kappa) coefficients
#        kappa in W/(m.K)
        self.kappa_param_1 = Param(initialize=0.118294,
                                   units=pyunits.W/(pyunits.m*pyunits.K),
                                   doc="Coefficient: thermal conductivity")
        self.kappa_param_2 = Param(initialize=-3.3E-5,
                                   units=pyunits.W/(pyunits.m*(pyunits.K**2)),
                                   doc="Coefficient: thermal conductivity")
        self.kappa_param_3 = Param(initialize=-1.5E-7,
                                   units=pyunits.W/(pyunits.m*(pyunits.K**3)),
                                   doc="Coefficient: thermal conductivity")

#        Thermodynamic reference state
#        ref_temperature in K
        self.ref_temperature = Param(within=PositiveReals,
                                     mutable=True,
                                     default=273.15,
                                     units=pyunits.K,
                                     doc='Reference temperature [K]')

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'flow_vol': {'method': None, 'units': 'm^3/s'},
                            'temperature': {'method': None, 'units': 'K'},
                            'pressure': {'method': None, 'units': 'Pa'},
                            'conc_mol_comp': {'method': None, 'units': 'mol/m^3'},
                            'dens_mol': {'method': None, 'units': 'mol/m^3'},
                            'cp_mass': {'method': None, 'units': 'J/kg/K'},
                            'enth_mass': {'method': None, 'units': 'J/kg'},
                            'dens_mass': {'method': None, 'units': 'kg/m^3'},
                            'visc_k_phase': {'method': None, 'units': 'm^2/s'},
                            'visc_d_phase': {'method': None,'units': 'Pa.s'},
                            'therm_cond_phase': {'method': None, 'units': 'W/m/K'}})
        obj.add_default_units({'time': pyunits.s,
                               'length': pyunits.m,
                               'mass': pyunits.kg,
                               'amount': pyunits.mol,
                               'temperature': pyunits.K})


class _StateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(blk, state_args=None,
                   outlvl=idaeslog.WARNING,
                   hold_state=False,
                   state_vars_fixed=False,
                   solver=None,
                   optarg=None):
        '''
        Initialization routine for property package.

        Keyword Arguments:
        state_args : Dictionary with initial guesses for the state vars
                     chosen. Note that if this method is triggered
                     through the control volume, and if initial guesses
                     were not provied at the unit model level, the
                     control volume passes the inlet values as initial
                     guess.The keys for the state_args dictionary are:

                     flow_mol_comp : value at which to initialize component
                                     flows (default=None)
                     pressure : value at which to initialize pressure
                                (default=None)
                     temperature : value at which to initialize temperature
                                  (default=None)
            outlvl : sets output level of initialization routine
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed and
                                       initialization does not need to worry
                                       about fixing and unfixing variables.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        '''

        if state_vars_fixed is False:
            # Fix state variables if not already fixed
            flags = fix_state_vars(blk, state_args)
        for k in blk.keys():
            if degrees_of_freedom(blk[k]) != 0:
                raise Exception("State vars fixed but degrees of freedom "
                                "for state block is not zero during "
                                "initialization.")

        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Solve property correlation
        results = solve_indexed_blocks(opt, [blk])
        if results.solver.termination_condition \
                == TerminationCondition.optimal:
            init_log.info('{} Initialisation Step 1 Complete.'.
                          format(blk.name))
        else:
            init_log.warning('{} Initialisation Step 1 Failed.'.
                             format(blk.name))

        init_log.info('Initialization Step 1 Complete.')

        # ---------------------------------------------------------------------
        if state_vars_fixed is False:
            # release state vars fixed during initialization if control
            # volume didn't fix the state vars
            if hold_state is True:
                return flags
            else:
                blk.release_state(flags)
        init_log.info('Initialization Complete.')

    def release_state(blk, flags, outlvl=idaeslog.NOTSET):
        # Method to release states only if explicitly called

        if flags is None:
            return
        # Unfix state variables
        revert_state_vars(blk, flags)
        init_log.info('State Released.')


@declare_process_block_class("ThermalOilStateBlock",
                             block_class=_StateBlock)
class ThermalOilStateBlockData(StateBlockData):
    """
    Property package for Thermal oil properties
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(ThermalOilStateBlockData, self).build()

        # Create state variables
        self._make_state_vars()

        # Create required properties
        self._make_prop_vars()

        # Create constraints
        self._make_constraints()

    def _make_state_vars(self):
        """Declare the necessary state variable objects."""
        self.flow_mass = Var(initialize=100.0,
                             domain=Reals,
                             doc="Total mass flow [Kg/s]",
                             units=pyunits.kg/pyunits.s)
        self.temperature = Var(initialize=523,
                               domain=Reals,
                               doc="Temperature of thermal oil [K]",
                               bounds=(260, 616),
                               units=pyunits.K)
        self.pressure = Var(initialize=101325,
                            domain=Reals,
                            doc="Pressure [Pa]",
                            units=pyunits.Pa)

    def _make_prop_vars(self):
        """Make additional variables for calculations."""

        self.enth_mass = Var(self.phase_list,
                             initialize=1,
                             units=pyunits.J/pyunits.kg,
                             doc='Specific Enthalpy')

    def _make_constraints(self):
        """Create property constraints."""

        # Specific heat capacity
        self.cp_mass = Expression(
            self.phase_list,
            expr=((self.params.cp_param_2 *
                   (self.temperature - self.params.ref_temperature) +
                   self.params.cp_param_3 *
                   (self.temperature - self.params.ref_temperature)**2 +
                   self.params.cp_param_1)),
            doc="Specific heat capacity [J/kg/K]"
        )

        # Specific Enthalpy
        def enthalpy_correlation(self, p):
            return (
                self.enth_mass[p]
                == ((self.params.cp_param_2 *
                     (self.temperature-self.params.ref_temperature)**2/2 +
                     self.params.cp_param_3 *
                     (self.temperature-self.params.ref_temperature)**3/3 +
                     self.params.cp_param_1*(self.temperature-self.params.ref_temperature))))
        self.enthalpy_eq = Constraint(self.phase_list,
                                      rule=enthalpy_correlation)

        # Kinematic viscosity
        self.visc_k_phase = Expression(
            self.phase_list,
            expr=(self.params.nu_param_4 * exp(
                self.params.nu_param_1 / (
                    self.temperature - self.params.ref_temperature +
                    self.params.nu_param_2) + self.params.nu_param_3)),
            doc="Kinematic viscosity [m2/s]"
        )

        # Thermal conductivity
        self.therm_cond_phase = Expression(
            self.phase_list,
            expr=(self.params.kappa_param_2 *
                  (self.temperature - self.params.ref_temperature) +
                  self.params.kappa_param_3 *
                  (self.temperature - self.params.ref_temperature)**2 +
                  self.params.kappa_param_1),
            doc="Thermal conductivity [W/m/K]"
        )

        # Density
        self.dens_mass = Expression(
            self.phase_list,
            expr=(self.params.rho_param_2 *
                  (self.temperature - self.params.ref_temperature) +
                  self.params.rho_param_1),
            doc="Density [kg/m3]"
        )

        # Dynamic viscosity
        def rule_visc_d_phase(b, p):
            return self.visc_k_phase[p] * self.dens_mass[p]

        self.visc_d_phase = Expression(
            self.config.parameters.phase_list,
            rule=rule_visc_d_phase)

        # Enthalpy flow terms
        def rule_enthalpy_flow_terms(b, p):
            return self.enth_mass[p] * self.flow_mass

        self.enthalpy_flow_terms = Expression(
            self.config.parameters.phase_list,
            rule=rule_enthalpy_flow_terms)

    def get_material_flow_terms(self, p, j):
        return self.flow_mass

    def get_enthalpy_flow_terms(self, p):
        return self.enthalpy_flow_terms[p]

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    def default_energy_balance_type(self):
        return EnergyBalanceType.enthalpyTotal

    def define_state_vars(self):
        """Define state variables for ports."""
        return {"flow_mass": self.flow_mass,
                "temperature": self.temperature,
                "pressure": self.pressure}

    def define_display_vars(self):
        return {"flow_mass": self.flow_mass,
                "temperature": self.temperature,
                "pressure": self.pressure}

    def get_material_flow_basis(self):
        return MaterialFlowBasis.mass
